import json
from typing import Optional
from unittest import IsolatedAsyncioTestCase

import pytest
from anoncreds import (
    AnoncredsError,
    Credential,
    CredentialDefinition,
    CredentialOffer,
    W3cCredential,
)
from aries_askar import AskarError, AskarErrorCode

from ...anoncreds.base import AnonCredsObjectAlreadyExists, AnonCredsSchemaAlreadyExists
from ...anoncreds.models.credential_definition import (
    CredDef,
    CredDefResult,
    CredDefState,
    CredDefValue,
    CredDefValuePrimary,
    CredDefValueRevocation,
    GetCredDefResult,
)
from ...anoncreds.models.schema import (
    AnonCredsSchema,
    GetSchemaResult,
    SchemaResult,
    SchemaState,
)
from ...askar.profile_anon import AskarAnonCredsProfile, AskarAnonCredsProfileSession
from ...core.event_bus import Event, MockEventBus
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import issuer as test_module


class MockSchemaEntry:
    def __init__(self, name: Optional[str] = "name"):
        self.name = name
        self.version = "1.0"
        self.value_json = "value_json"
        self.issuer_id = "issuer-id"

    def serialize(self):
        return self.value_json


class MockSchemaValue:
    attr_names = ["attr1", "attr2"]


class MockSchemaResult:
    def __init__(self) -> None:
        self.schema = MockSchemaEntry()
        self.schema_value = MockSchemaValue()


class MockCredDefState:
    credential_definition_id = "cred-def-id"
    state = "finished"


class MockCredDefEntry:
    def __init__(self, name: Optional[str] = "name", epoch: Optional[str] = None):
        self.name = name
        self.tags = {
            "schema_id": "schema-id",
            "epoch": epoch,
        }

    credential_definition_state = MockCredDefState()
    raw_value = "raw-value"
    value_json = {}

    def to_json(self):
        return json.dumps({"cred_def": "cred_def"})


class MockCredDefPrivate:
    def to_json_buffer(self):
        return "cred-def-private"


class MockKeyProof:
    def to_json_buffer(self):
        return "key-proof"

    raw_value = "raw-value"


class MockCredOffer:
    def to_json(self):
        return json.dumps({"cred_offer": "cred_offer"})


class MockCredential:
    def to_json(self):
        return json.dumps({"credential": "credential"})


def get_mock_schema_result(
    job_id: Optional[str] = "job-id", schema_id: Optional[str] = "schema-id"
):
    return SchemaResult(
        job_id=job_id,
        schema_state=SchemaState(
            state="finished",
            schema_id=schema_id,
            schema=AnonCredsSchema(
                issuer_id="issuer-id",
                name="name",
                version="1.0",
                attr_names=["attr1", "attr2"],
            ),
        ),
    )


@pytest.mark.anoncreds
class TestAnonCredsIssuer(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.profile = await create_test_profile(
            settings={"wallet.type": "askar-anoncreds"},
        )
        self.issuer = test_module.AnonCredsIssuer(self.profile)

    async def test_init(self):
        assert isinstance(self.issuer, test_module.AnonCredsIssuer)
        assert isinstance(self.issuer.profile, AskarAnonCredsProfile)

    async def test_init_wrong_profile_type(self):
        self.issuer._profile = await create_test_profile(
            settings={"wallet.type": "askar"},
        )
        with self.assertRaises(ValueError):
            self.issuer.profile

    async def test_notify(self):
        self.profile.inject = mock.Mock(return_value=MockEventBus())
        await self.issuer.notify(Event(topic="test-topic"))

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(AnonCredsSchema, "deserialize", return_value="test")
    async def test_create_and_register_schema_finds_schema_raises_x(
        self, _, mock_session_handle
    ):
        mock_schema = AnonCredsSchema(
            issuer_id="issuer-id",
            name="schema-name",
            version="1.0",
            attr_names=["attr1", "attr2"],
        )
        mock_schema.value_json = "value_json"
        mock_session_handle.fetch_all = mock.CoroutineMock(return_value=[mock_schema])
        with self.assertRaises(AnonCredsObjectAlreadyExists):
            await self.issuer.create_and_register_schema(
                issuer_id="issuer-id",
                name="name",
                version="1.0",
                attr_names=["attr1", "attr2"],
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_create_and_register_schema(self, mock_session_handle):
        mock_session_handle.fetch_all = mock.CoroutineMock(return_value=[])
        mock_session_handle.insert = mock.CoroutineMock(return_value=None)
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                register_schema=mock.CoroutineMock(return_value=get_mock_schema_result())
            )
        )
        result = await self.issuer.create_and_register_schema(
            issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
            name="example name",
            version="1.0",
            attr_names=["attr1", "attr2"],
        )

        assert result is not None
        mock_session_handle.fetch_all.assert_called_once()
        mock_session_handle.insert.assert_called_once()

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_create_and_register_schema_missing_schema_id_or_job_id(
        self, mock_session_handle
    ):
        mock_session_handle.fetch_all = mock.CoroutineMock(return_value=[])
        mock_session_handle.insert = mock.CoroutineMock(return_value=None)
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                register_schema=mock.CoroutineMock(
                    side_effect=[
                        SchemaResult(
                            job_id=None,
                            schema_state=SchemaState(
                                state="finished",
                                schema_id=None,
                                schema=AnonCredsSchema(
                                    issuer_id="issuer-id",
                                    name="name",
                                    version="1.0",
                                    attr_names=["attr1", "attr2"],
                                ),
                            ),
                        ),
                        SchemaResult(
                            job_id=None,
                            schema_state=SchemaState(
                                state="finished",
                                schema_id="schema-id",
                                schema=AnonCredsSchema(
                                    issuer_id="issuer-id",
                                    name="name",
                                    version="1.0",
                                    attr_names=["attr1", "attr2"],
                                ),
                            ),
                        ),
                        SchemaResult(
                            job_id="job-id",
                            schema_state=SchemaState(
                                state="finished",
                                schema_id=None,
                                schema=AnonCredsSchema(
                                    issuer_id="issuer-id",
                                    name="name",
                                    version="1.0",
                                    attr_names=["attr1", "attr2"],
                                ),
                            ),
                        ),
                    ]
                )
            )
        )

        with self.assertRaises(ValueError):
            await self.issuer.create_and_register_schema(
                issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                name="example name",
                version="1.0",
                attr_names=["attr1", "attr2"],
            )

            mock_session_handle.fetch_all.assert_called_once()
            mock_session_handle.insert.assert_not_called

        await self.issuer.create_and_register_schema(
            issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
            name="example name",
            version="1.0",
            attr_names=["attr1", "attr2"],
        )
        await self.issuer.create_and_register_schema(
            issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
            name="example name",
            version="1.0",
            attr_names=["attr1", "attr2"],
        )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_create_and_register_schema_fail_insert(self, mock_session_handle):
        mock_session_handle.fetch_all = mock.CoroutineMock(return_value=[])
        mock_session_handle.insert = mock.CoroutineMock(
            side_effect=AskarError(AskarErrorCode.UNEXPECTED, message="test-msg")
        )
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                register_schema=mock.CoroutineMock(return_value=get_mock_schema_result())
            )
        )

        with self.assertRaises(test_module.AnonCredsIssuerError):
            result = await self.issuer.create_and_register_schema(
                issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                name="example name",
                version="1.0",
                attr_names=["attr1", "attr2"],
            )

            assert result is not None
            mock_session_handle.fetch_all.assert_called_once()
            mock_session_handle.insert.assert_called_once()

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_create_and_register_schema_already_exists_but_not_in_wallet(
        self, mock_session_handle
    ):
        mock_session_handle.fetch_all = mock.CoroutineMock(return_value=[])
        mock_session_handle.insert = mock.CoroutineMock(return_value=None)
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                register_schema=mock.CoroutineMock(
                    side_effect=AnonCredsSchemaAlreadyExists(
                        message="message",
                        obj_id="id",
                        obj=AnonCredsSchema(
                            issuer_id="issuer-id",
                            name="schema-name",
                            version="1.0",
                            attr_names=["attr1", "attr2"],
                        ),
                    )
                )
            )
        )
        with self.assertRaises(test_module.AnonCredsIssuerError):
            await self.issuer.create_and_register_schema(
                issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                name="example",
                version="1.0",
                attr_names=["attr1", "attr2"],
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_create_and_register_schema_without_job_id_or_schema_id_raises_x(
        self, mock_session_handle
    ):
        mock_session_handle.fetch_all = mock.CoroutineMock(return_value=[])
        mock_session_handle.insert = mock.CoroutineMock(return_value=None)
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                register_schema=mock.CoroutineMock(
                    side_effect=[
                        get_mock_schema_result(job_id=None, schema_id=None),
                        get_mock_schema_result(job_id=None),
                        get_mock_schema_result(schema_id=None),
                    ]
                )
            )
        )
        with self.assertRaises(ValueError):
            await self.issuer.create_and_register_schema(
                issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                name="name",
                version="1.0",
                attr_names=["attr1", "attr2"],
            )
        await self.issuer.create_and_register_schema(
            issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
            name="name",
            version="1.0",
            attr_names=["attr1", "attr2"],
        )
        await self.issuer.create_and_register_schema(
            issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
            name="name",
            version="1.0",
            attr_names=["attr1", "attr2"],
        )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(test_module.AnonCredsIssuer, "store_schema")
    async def test_create_and_register_schema_with_endorsed_transaction_response_does_not_store_schema(
        self, mock_store_schema, mock_session_handle
    ):
        mock_session_handle.fetch_all = mock.CoroutineMock(return_value=[])
        mock_session_handle.insert = mock.CoroutineMock(return_value=None)
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                register_schema=mock.CoroutineMock(
                    return_value=SchemaResult(
                        job_id="job-id",
                        schema_state=SchemaState(
                            state="finished",
                            schema_id="schema-id",
                            schema=AnonCredsSchema(
                                issuer_id="issuer-id",
                                name="schema-name",
                                version="1.0",
                                attr_names=["attr1", "attr2"],
                            ),
                        ),
                    )
                )
            )
        )
        result = await self.issuer.create_and_register_schema(
            issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
            name="example name",
            version="1.0",
            attr_names=["attr1", "attr2"],
        )

        assert isinstance(result, SchemaResult)
        assert mock_store_schema.called

    async def test_finish_schema(self):
        self.profile.transaction = mock.Mock(
            return_value=mock.MagicMock(
                commit=mock.CoroutineMock(return_value=None),
            )
        )
        await self.issuer.finish_schema(job_id="job-id", schema_id="schema-id")

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_created_schemas(self, mock_session_handle):
        mock_session_handle.fetch_all = mock.CoroutineMock(
            return_value=[MockSchemaEntry("name-test")]
        )
        result = await self.issuer.get_created_schemas()
        mock_session_handle.fetch_all.assert_called_once()
        assert result == ["name-test"]

        mock_session_handle.fetch_all = mock.CoroutineMock(
            return_value=[
                MockSchemaEntry("schema1"),
                MockSchemaEntry("schema2"),
            ]
        )
        result = await self.issuer.get_created_schemas()
        mock_session_handle.fetch_all.assert_called_once()
        assert result == ["schema1", "schema2"]

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_credential_definition_in_wallet(self, mock_session_handle):
        mock_session_handle.fetch = mock.CoroutineMock(
            side_effect=[
                CredDef(
                    issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                    schema_id="schema-id",
                    tag="tag",
                    type="CL",
                    value=CredDefValue(
                        primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                    ),
                ),
                None,
                AskarError(AskarErrorCode.UNEXPECTED, message="test-msg"),
            ]
        )
        assert await self.issuer.credential_definition_in_wallet("cred-def-id") is True
        assert await self.issuer.credential_definition_in_wallet("cred-def-id") is False
        with self.assertRaises(test_module.AnonCredsIssuerError):
            await self.issuer.credential_definition_in_wallet("cred-def-id")

    async def test_create_and_register_credential_definition_invalid_options_raises_x(
        self,
    ):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(
                    return_value=AnonCredsSchema(
                        issuer_id="issuer-id",
                        name="schema-name",
                        version="1.0",
                        attr_names=["attr1", "attr2"],
                    )
                )
            )
        )
        with self.assertRaises(ValueError):
            await self.issuer.create_and_register_credential_definition(
                issuer_id="issuer-id",
                schema_id="schema-id",
                signature_type="CL",
                options={"support_revocation": "true"},  # requires boolean
            )
        with self.assertRaises(ValueError):
            await self.issuer.create_and_register_credential_definition(
                issuer_id="issuer-id",
                schema_id="schema-id",
                signature_type="CL",
                options={"revocation_registry_size": "100"},  # requires integer
            )

    @mock.patch.object(CredDef, "from_native", return_value=MockCredDefEntry())
    @mock.patch(
        "anoncreds.CredentialDefinition.create",
        return_value=(mock.MagicMock(), mock.MagicMock(), mock.MagicMock()),
    )
    async def test_create_and_register_credential_definition_support_revocation_conditions(
        self, mock_cred_def_create, _
    ):
        schema_result = GetSchemaResult(
            schema_id="schema-id",
            schema=AnonCredsSchema(
                issuer_id="issuer-id",
                name="schema-name",
                version="1.0",
                attr_names=["attr1", "attr2"],
            ),
            schema_metadata={},
            resolution_metadata={},
        )

        cred_def_result = CredDefResult(
            job_id="job-id",
            credential_definition_state=CredDefState(
                state="finished",
                credential_definition=CredDef(
                    issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                    schema_id="schema-id",
                    tag="tag",
                    type="CL",
                    value=CredDefValue(
                        primary=CredDefValuePrimary("n", "s", {}, "rctxt", "z")
                    ),
                ),
                credential_definition_id="cred-def-id",
            ),
            credential_definition_metadata={},
            registration_metadata={},
        )

        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(return_value=schema_result),
                register_credential_definition=mock.CoroutineMock(
                    return_value=cred_def_result
                ),
            )
        )

        # Configure author role and auto create rev reg -- expectation: support revocation is True when not specified
        self.profile.settings.set_value("endorser.author", True)
        self.profile.settings.set_value("endorser.auto_create_rev_reg", True)

        # First assert AnonCredsIssuerError if tails_server_base_url is not set
        with self.assertRaises(test_module.AnonCredsIssuerError) as exc:
            await self.issuer.create_and_register_credential_definition(
                issuer_id="issuer-id",
                schema_id="schema-id",
                signature_type="CL",
                tag="tag",
            )
        assert (
            str(exc.exception.message)
            == "tails_server_base_url not configured. Can't create revocable credential definition."
        )

        # Now, set the tails_server_base_url
        self.profile.settings.set_value("tails_server_base_url", "https://example.com")

        for support_revocation in [True, False, None]:
            # Mock the store_credential_definition method
            with mock.patch.object(
                self.issuer, "store_credential_definition"
            ) as mock_store_cred_def:
                # Reset the mocks for each iteration
                mock_cred_def_create.reset_mock()
                mock_store_cred_def.reset_mock()

                await self.issuer.create_and_register_credential_definition(
                    issuer_id="issuer-id",
                    schema_id="schema-id",
                    signature_type="CL",
                    tag="tag",
                    options={"support_revocation": support_revocation},
                )

                # Check if support_revocation is True when None or True was passed
                expected_support_revocation = (
                    support_revocation if support_revocation is not None else True
                )

                # Assert CredentialDefinition.create call was made with correct support_revocation value
                mock_cred_def_create.assert_called_once_with(
                    schema_id="schema-id",
                    schema=schema_result.schema.serialize(),
                    issuer_id="issuer-id",
                    tag="tag",
                    signature_type="CL",
                    support_revocation=expected_support_revocation,
                )

                # Assert store_credential_definition call args
                mock_store_cred_def.assert_called_once_with(
                    schema_result=schema_result,
                    cred_def_result=mock.ANY,
                    cred_def_private=mock.ANY,
                    key_proof=mock.ANY,
                    support_revocation=expected_support_revocation,
                    max_cred_num=mock.ANY,
                    options=mock.ANY,
                )

        # Now, disable author role and auto create rev reg -- expectation: support revocation is False when not specified
        self.profile.settings.set_value("endorser.author", False)
        self.profile.settings.set_value("endorser.auto_create_rev_reg", False)

        for support_revocation in [True, False, None]:
            # Mock the CredentialDefinition.create call, and the store_credential_definition method
            with mock.patch.object(
                self.issuer, "store_credential_definition"
            ) as mock_store_cred_def:
                # Reset the mock for each iteration
                mock_cred_def_create.reset_mock()
                mock_store_cred_def.reset_mock()

                await self.issuer.create_and_register_credential_definition(
                    issuer_id="issuer-id",
                    schema_id="schema-id",
                    signature_type="CL",
                    tag="tag",
                    options={"support_revocation": support_revocation},
                )

                # Check if support_revocation is False when set to None
                expected_support_revocation = (
                    support_revocation if support_revocation is not None else False
                )

                # Assert CredentialDefinition.create call was made with correct support_revocation value
                mock_cred_def_create.assert_called_once_with(
                    schema_id="schema-id",
                    schema=schema_result.schema.serialize(),
                    issuer_id="issuer-id",
                    tag="tag",
                    signature_type="CL",
                    support_revocation=expected_support_revocation,
                )

                # Assert store_credential_definition call args
                mock_store_cred_def.assert_called_once_with(
                    schema_result=schema_result,
                    cred_def_result=mock.ANY,
                    cred_def_private=mock.ANY,
                    key_proof=mock.ANY,
                    support_revocation=expected_support_revocation,
                    max_cred_num=mock.ANY,
                    options=mock.ANY,
                )

    @mock.patch.object(test_module.AnonCredsIssuer, "notify")
    async def test_create_and_register_credential_definition_finishes(self, mock_notify):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(
                    return_value=GetSchemaResult(
                        schema_id="schema-id",
                        schema=AnonCredsSchema(
                            issuer_id="issuer-id",
                            name="schema-name",
                            version="1.0",
                            attr_names=["attr1", "attr2"],
                        ),
                        schema_metadata={},
                        resolution_metadata={},
                    )
                ),
                register_credential_definition=mock.CoroutineMock(
                    return_value=CredDefResult(
                        job_id="job-id",
                        credential_definition_state=CredDefState(
                            state="finished",
                            credential_definition=CredDef(
                                issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                                schema_id="schema-id",
                                tag="tag",
                                type="CL",
                                value=CredDefValue(
                                    primary=CredDefValuePrimary(
                                        "n", "s", {}, "rctxt", "z"
                                    )
                                ),
                            ),
                            credential_definition_id="cred-def-id",
                        ),
                        credential_definition_metadata={},
                        registration_metadata={},
                    )
                ),
            )
        )
        self.profile.transaction = mock.Mock(
            return_value=mock.MagicMock(
                insert=mock.CoroutineMock(return_value=None),
                commit=mock.CoroutineMock(return_value=None),
            )
        )
        result = await self.issuer.create_and_register_credential_definition(
            issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
            schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
            signature_type="CL",
            options={},
        )

        assert isinstance(result, CredDefResult)
        mock_notify.assert_called_once()

    @mock.patch.object(test_module.AnonCredsIssuer, "notify")
    async def test_create_and_register_credential_definition_errors(self, mock_notify):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(
                    return_value=GetSchemaResult(
                        schema_id="schema-id",
                        schema=AnonCredsSchema(
                            issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                            name="schema-name",
                            version="1.0",
                            attr_names=["attr1", "attr2"],
                        ),
                        schema_metadata={},
                        resolution_metadata={},
                    )
                ),
                # No job_id or cred_def_id
                register_credential_definition=mock.CoroutineMock(
                    side_effect=[
                        CredDefResult(
                            job_id=None,
                            credential_definition_state=CredDefState(
                                state="finished",
                                credential_definition=CredDef(
                                    issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                                    schema_id="schema-id",
                                    tag="tag",
                                    type="CL",
                                    value=CredDefValue(
                                        primary=CredDefValuePrimary(
                                            "n", "s", {}, "rctxt", "z"
                                        )
                                    ),
                                ),
                                credential_definition_id=None,
                            ),
                            credential_definition_metadata={},
                            registration_metadata={},
                        ),
                    ]
                ),
            )
        )
        self.profile.transaction = mock.Mock(
            return_value=mock.MagicMock(
                insert=mock.CoroutineMock(return_value=None),
                commit=mock.CoroutineMock(return_value=None),
            )
        )
        # Creating fails with bad issuer_id
        with self.assertRaises(AnoncredsError):
            await self.issuer.create_and_register_credential_definition(
                issuer_id="issuer-id",
                schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                signature_type="CL",
                options={},
            )
        # No job_id or cred_def_id
        with self.assertRaises(test_module.AnonCredsIssuerError):
            await self.issuer.create_and_register_credential_definition(
                issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                schema_id="CsQY9MGeD3CQP4EyuVFo5m:2:MYCO Biomarker:0.0.3",
                signature_type="CL",
                options={},
            )

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_get_created_cred_defs(self, mock_session_handle):
        mock_session_handle.fetch_all = mock.CoroutineMock(
            return_value=[MockCredDefEntry()]
        )
        result = await self.issuer.get_created_credential_definitions()
        mock_session_handle.fetch_all.assert_called_once()
        assert result == ["name"]
        mock_session_handle.fetch_all = mock.CoroutineMock(
            return_value=[MockCredDefEntry("cred_def1"), MockCredDefEntry("cred_def2")]
        )
        result = await self.issuer.get_created_credential_definitions(
            issuer_id="issuer-id-test",
            schema_issuer_id="schema-issuer-id-test",
            schema_id="schema-id-test",
            schema_name="schema-name-test",
            schema_version="schema-version-test",
            epoch="123456789",
        )
        mock_session_handle.fetch_all.assert_called_once()
        assert result == ["cred_def1", "cred_def2"]

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_match_created_cred_defs(self, mock_session_handle):
        mock_session_handle.fetch_all = mock.CoroutineMock(
            return_value=[
                MockCredDefEntry(name="name2", epoch="2"),
                MockCredDefEntry(name="name3", epoch="3"),
                MockCredDefEntry(name="name4", epoch="4"),
                MockCredDefEntry(name="name1", epoch="1"),
            ]
        )
        result = await self.issuer.match_created_credential_definitions()
        assert result == "name4"

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    async def test_create_credential_offer_cred_def_not_found(self, mock_session_handle):
        """
        None, Valid
        Valid, None
        None, None
        """
        mock_session_handle.fetch = mock.CoroutineMock(
            side_effect=[None, MockKeyProof(), MockCredDefEntry(), None, None, None]
        )
        with self.assertRaises(test_module.AnonCredsIssuerError):
            await self.issuer.create_credential_offer("cred-def-id")
        with self.assertRaises(test_module.AnonCredsIssuerError):
            await self.issuer.create_credential_offer("cred-def-id")
        with self.assertRaises(test_module.AnonCredsIssuerError):
            await self.issuer.create_credential_offer("cred-def-id")

    async def test_cred_def_supports_revocation(self):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_credential_definition=mock.CoroutineMock(
                    side_effect=[
                        GetCredDefResult(
                            credential_definition_id="cred-def-id",
                            credential_definition=CredDef(
                                issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                                schema_id="schema-id",
                                tag="tag",
                                type="CL",
                                value=CredDefValue(
                                    primary=CredDefValuePrimary(
                                        "n", "s", {}, "rctxt", "z"
                                    )
                                ),
                            ),
                            credential_definition_metadata={},
                            resolution_metadata={},
                        ),
                        GetCredDefResult(
                            credential_definition_id="cred-def-id",
                            credential_definition=CredDef(
                                issuer_id="did:sov:3avoBCqDMFHFaKUHug9s8W",
                                schema_id="schema-id",
                                tag="tag",
                                type="CL",
                                value=CredDefValue(
                                    primary=CredDefValuePrimary(
                                        "n", "s", {}, "rctxt", "z"
                                    ),
                                    revocation=CredDefValueRevocation(
                                        g="g",
                                        g_dash="g_dash",
                                        h="h",
                                        h0="h0",
                                        h1="h1",
                                        h2="h2",
                                        h_cap="h_cap",
                                        htilde="htilde",
                                        pk="pk",
                                        u="u",
                                        y="y",
                                    ),
                                ),
                            ),
                            credential_definition_metadata={},
                            resolution_metadata={},
                        ),
                    ]
                )
            )
        )

        result = await self.issuer.cred_def_supports_revocation("cred-def-id")
        assert result is False
        result = await self.issuer.cred_def_supports_revocation("cred-def-id")
        assert result is True

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(CredentialDefinition, "load", return_value=MockCredDefEntry())
    async def test_create_credential_offer_create_fail(
        self, mock_load, mock_session_handle
    ):
        mock_session_handle.fetch = mock.CoroutineMock(
            side_effect=[MockCredDefEntry(), MockKeyProof()]
        )
        with self.assertRaises(test_module.AnonCredsIssuerError):
            await self.issuer.create_credential_offer("cred-def-id")
        assert mock_session_handle.fetch.called
        assert mock_load.called

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(CredentialDefinition, "load", return_value=MockCredDefEntry())
    @mock.patch.object(CredentialOffer, "create", return_value=MockCredOffer())
    async def test_create_credential_offer_create(
        self, mock_create, mock_load, mock_session_handle
    ):
        mock_session_handle.fetch = mock.CoroutineMock(
            side_effect=[MockCredDefEntry(), MockKeyProof()]
        )
        result = await self.issuer.create_credential_offer("cred-def-id")
        assert mock_session_handle.fetch.called
        assert mock_load.called
        assert mock_create.called
        assert result is not None

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(Credential, "create", return_value=MockCredential())
    async def test_create_credential(self, mock_create, mock_session_handle):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(return_value=MockSchemaResult())
            )
        )
        mock_session_handle.fetch = mock.CoroutineMock(return_value=MockCredDefEntry())
        result = await self.issuer.create_credential(
            {"schema_id": "schema-id", "cred_def_id": "cred-def-id"},
            {},
            {"attr1": "value1", "attr2": "value2"},
        )

        assert result is not None
        assert mock_session_handle.fetch.called
        assert mock_create.called

    @mock.patch.object(AskarAnonCredsProfileSession, "handle")
    @mock.patch.object(W3cCredential, "create", return_value=MockCredential())
    async def test_create_credential_vcdi(self, mock_create, mock_session_handle):
        self.profile.inject = mock.Mock(
            return_value=mock.MagicMock(
                get_schema=mock.CoroutineMock(return_value=MockSchemaResult()),
            )
        )

        mock_session_handle.fetch = mock.CoroutineMock(return_value=MockCredDefEntry())
        result = await self.issuer.create_credential_w3c(
            {"schema_id": "schema-id", "cred_def_id": "cred-def-id"},
            {},
            {"attr1": "value1", "attr2": "value2"},
        )

        assert result is not None
        assert mock_session_handle.fetch.called
        assert mock_create.called



TEST_LD_DOCUMENT = {
    "@context": ["http://schema.org/", "https://w3id.org/security/bbs/v1"],
    "@type": "Person",
    "firstName": "Jane",
    "lastName": "Does",
    "jobTitle": "Professor",
    "telephone": "(425) 123-4567",
    "email": "jane.doe@example.com",
}



"""Class to represent a Linked Data proof set."""

from typing import List, Optional, Union

from pyld.jsonld import JsonLdProcessor

from .constants import SECURITY_CONTEXT_URL
from .document_loader import DocumentLoaderMethod
from .error import LinkedDataProofException
from .purposes.proof_purpose import ProofPurpose
from .suites import _LinkedDataProof as LinkedDataProof
from .validation_result import DocumentVerificationResult, ProofResult


class ProofSet:
    """Class for managing proof sets on a JSON-LD document."""

    @staticmethod
    async def add(
        *,
        document: dict,
        suite: LinkedDataProof,
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> dict:
        """Add a Linked Data proof to the document.

        If the document contains other proofs, the proof will be appended
        to the existing set of proofs.

        Important note: This method assumes that the term `proof` in the given
        document has the same definition as the `https://w3id.org/security/v2`
        JSON-LD @context.

        Args:
            document (dict): JSON-LD document to be signed.
            suite (LinkedDataProof): A signature suite instance that will create the proof
            purpose (ProofPurpose): A proof purpose instance that will augment the proof
                with information describing its intended purpose.
            document_loader (DocumentLoader): Document loader to use.

        Returns:
            dict: The signed document, with the signature in the top-level
                `proof` property.

        """
        # Shallow copy document to allow removal of existing proofs
        input = document.copy()
        input.pop("proof", None)

        # create the new proof, suites MUST output a proof using security-v2 `@context`
        proof = await suite.create_proof(
            document=input, purpose=purpose, document_loader=document_loader
        )

        JsonLdProcessor.add_value(document, "proof", proof)
        return document

    @staticmethod
    async def verify(
        *,
        document: dict,
        suites: List[LinkedDataProof],
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> DocumentVerificationResult:
        """Verify Linked Data proof(s) on a document.

        The proofs to be verified must match the given proof purse.

        Important note: This method assumes that the term `proof` in the given
        document has the same definition as the `https://w3id.org/security/v2`
        JSON-LD @context.

        Args:
            document (dict): JSON-LD document with one or more proofs to be verified.
            suites (List[LinkedDataProof]): Acceptable signature suite instances
                for verifying the proof(s).
            purpose (ProofPurpose): A proof purpose instance that will match proofs
                to be verified and ensure they were created according to the
                appropriate purpose.
            document_loader (DocumentLoader): Document loader to use.

        Returns:
            DocumentVerificationResult: Object with a `verified` property that is `true`
                if at least one proof matching the given purpose and suite verifies
                and `false` otherwise. Also contains `errors` and `results` properties
                with extra data.

        """
        try:
            # Shallow copy document to allow removal of proof property without
            # modifying external document
            input = document.copy()

            if len(suites) == 0:
                raise LinkedDataProofException("At least one suite is required.")

            # Get proofs from document, remove proof property
            proof_set = await ProofSet._get_proofs(document=input)
            input.pop("proof", None)

            results = await ProofSet._verify(
                document=input,
                suites=suites,
                proof_set=proof_set,
                purpose=purpose,
                document_loader=document_loader,
            )

            # If no proofs were verified because of no matching suites and purposes
            # throw an error
            if len(results) == 0:
                suite_names = ", ".join([suite.signature_type for suite in suites])
                raise LinkedDataProofException(
                    f"Could not verify any proofs; no proofs matched the required"
                    f" suites ({suite_names}) and purpose ({purpose.term})"
                )

            # check if all results are valid, create result
            verified = any(result.verified for result in results)
            result = DocumentVerificationResult(
                verified=verified, document=document, results=results
            )

            # If not valid, extract and optionally add errors to result
            if not verified:
                errors = [result.error for result in results if result.error]

                if len(errors) > 0:
                    result.errors = errors

            return result
        except Exception as e:
            return DocumentVerificationResult(
                verified=False, document=document, errors=[e]
            )

    @staticmethod
    async def derive(
        *,
        document: dict,
        reveal_document: dict,
        # TODO: I think this could support multiple suites?
        # But then, why do multiple proofs?
        suite: LinkedDataProof,
        document_loader: DocumentLoaderMethod,
        nonce: Optional[bytes] = None,
    ) -> dict:
        """Create new derived Linked Data proof(s) on document using the reveal document.

        Important note: This method assumes that the term `proof` in the given
        document has the same definition as the `https://w3id.org/security/v2`
        JSON-LD @context. (v3 because BBS?)

        Args:
            document (dict): JSON-LD document with one or more proofs to be derived.
            reveal_document (dict): JSON-LD frame specifying the attributes to reveal.
            suite (LinkedDataProof): A signature suite instance to derive the proof.
            document_loader (DocumentLoader): Document loader to use.
            nonce (bytes, optional): Nonce to use for the proof. Defaults to None.

        Returns:
            dict: The derived document with the derived proof(s) in the top-level
                `proof` property.

        """
        # Shallow copy document to allow removal of existing proofs
        input = document.copy()

        # Check if suite supports derivation
        if not suite.supported_derive_proof_types:
            raise LinkedDataProofException(
                f"{suite.signature_type} does not support derivation"
            )

        # Get proofs, remove proof from document
        proof_set = await ProofSet._get_proofs(
            document=input, proof_types=suite.supported_derive_proof_types
        )
        input.pop("proof", None)

        # Derive proof, remove context
        derived_proof = await suite.derive_proof(
            proof=proof_set[0],
            document=input,
            reveal_document=reveal_document,
            document_loader=document_loader,
            nonce=nonce,
        )

        if len(proof_set) > 1:
            derived_proof["proof"] = [derived_proof["proof"]]

            proof_set.pop(0)

            for proof in proof_set:
                additional_derived_proof = await suite.derive_proof(
                    proof=proof,
                    document=input,
                    reveal_document=reveal_document,
                    document_loader=document_loader,
                )
                derived_proof["proof"].append(additional_derived_proof["proof"])

        JsonLdProcessor.add_value(
            derived_proof["document"], "proof", derived_proof["proof"]
        )

        return derived_proof["document"]

    @staticmethod
    async def _get_proofs(
        document: dict, proof_types: Union[List[str], None] = None
    ) -> list:
        """Get proof set from document, optionally filtered by proof_types."""
        proof_set = JsonLdProcessor.get_values(document, "proof")

        # If proof_types is present, only take proofs that match
        if proof_types:
            proof_set = list(filter(lambda _: _["type"] in proof_types, proof_set))

        if len(proof_set) == 0:
            raise LinkedDataProofException(
                "No matching proofs found in the given document"
            )

        # Shallow copy proofs and add document context or SECURITY_CONTEXT_URL
        context = document.get("@context") or SECURITY_CONTEXT_URL
        proof_set = [{"@context": context, **proof} for proof in proof_set]

        return proof_set

    @staticmethod
    async def _verify(
        document: dict,
        suites: List[LinkedDataProof],
        proof_set: List[dict],
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> List[ProofResult]:
        """Verify proofs in proof set.

        Returns results for proofs that match both on purpose and have a suite
        in the suites lists. This means proofs that don't match on any of these
        WILL NOT be verified OR included in the proof result list.
        """

        # Matches proof purposes proof set to passed purpose.
        # Only proofs with a `proofPurpose` that match the purpose are verified
        # e.g.:
        #   purpose = {term = 'assertionMethod'}
        #   proof_set = [{proofPurpose:'assertionMethod'},{proofPurpose: 'another'}]
        #   return = [ { proofPurpose: 'assertionMethod' } ]
        matches = [proof for proof in proof_set if purpose.match(proof)]

        if len(matches) == 0:
            return []

        results = []

        for proof in matches:
            for suite in suites:
                if suite.match_proof(proof.get("type")):
                    result = await suite.verify_proof(
                        proof=proof,
                        document=document,
                        purpose=purpose,
                        document_loader=document_loader,
                    )
                    result.proof = proof

                    results.append(result)

        return results



from .check import get_properties_without_context
from .crypto import _KeyPair as KeyPair
from .crypto import _WalletKeyPair as WalletKeyPair
from .document_loader import DocumentLoader, DocumentLoaderMethod
from .error import LinkedDataProofException
from .ld_proofs import derive, sign, verify
from .proof_set import ProofSet
from .purposes import _AssertionProofPurpose as AssertionProofPurpose
from .purposes import _AuthenticationProofPurpose as AuthenticationProofPurpose
from .purposes import _ControllerProofPurpose as ControllerProofPurpose
from .purposes import _CredentialIssuancePurpose as CredentialIssuancePurpose
from .purposes import _ProofPurpose as ProofPurpose
from .suites import _BbsBlsSignature2020 as BbsBlsSignature2020
from .suites import _BbsBlsSignatureProof2020 as BbsBlsSignatureProof2020
from .suites import _EcdsaSecp256r1Signature2019 as EcdsaSecp256r1Signature2019
from .suites import _Ed25519Signature2018 as Ed25519Signature2018
from .suites import _Ed25519Signature2020 as Ed25519Signature2020
from .suites import _JwsLinkedDataSignature as JwsLinkedDataSignature
from .suites import _LinkedDataProof as LinkedDataProof
from .suites import _LinkedDataSignature as LinkedDataSignature
from .validation_result import DocumentVerificationResult, ProofResult, PurposeResult

__all__ = [
    "sign",
    "verify",
    "derive",
    "ProofSet",
    # Proof purposes
    "ProofPurpose",
    "ControllerProofPurpose",
    "AssertionProofPurpose",
    "AuthenticationProofPurpose",
    "CredentialIssuancePurpose",
    # Suites
    "LinkedDataProof",
    "LinkedDataSignature",
    "JwsLinkedDataSignature",
    "Ed25519Signature2018",
    "Ed25519Signature2020",
    "EcdsaSecp256r1Signature2019",
    "BbsBlsSignature2020",
    "BbsBlsSignatureProof2020",
    # Key pairs
    "KeyPair",
    "WalletKeyPair",
    # Document Loaders
    "DocumentLoaderMethod",
    "DocumentLoader",
    # Exceptions
    "LinkedDataProofException",
    # Validation results
    "DocumentVerificationResult",
    "ProofResult",
    "PurposeResult",
    "get_properties_without_context",
]



from unittest import TestCase

from ......messaging.decorators.attach_decorator import AttachDecorator
from ..cred_format import V20CredFormat
from ..inner.cred_preview import V20CredAttrSpec, V20CredPreview

TEST_PREVIEW = V20CredPreview(
    attributes=(
        V20CredAttrSpec.list_plain({"test": "123", "hello": "world"})
        + [V20CredAttrSpec(name="icon", value="cG90YXRv", mime_type="image/png")]
    )
)

TEST_INDY_FILTER = {
    "schema_id": "GMm4vMw8LLrLJjp81kRRLp:2:ahoy:1560364003.0",
    "cred_def_id": "GMm4vMw8LLrLJjp81kRRLp:3:CL:12:tag",
}


class TestV20FormatFormat(TestCase):
    """Coverage for self-get."""

    def test_get_completeness(self):
        assert (
            V20CredFormat.Format.get(V20CredFormat.Format.INDY)
            is V20CredFormat.Format.INDY
        )
        assert V20CredFormat.Format.get("no such format") is None
        assert V20CredFormat.Format.get("hlindy/...") is V20CredFormat.Format.INDY
        assert (
            V20CredFormat.Format.get("aries/...").detail.__name__
            == "V20CredExRecordLDProof"
        )
        assert (
            V20CredFormat.Format.get(V20CredFormat.Format.LD_PROOF.api)
            is V20CredFormat.Format.LD_PROOF
        )

    def test_get_attachment_data(self):
        assert (
            V20CredFormat.Format.INDY.get_attachment_data(
                formats=[
                    V20CredFormat(attach_id="indy", format_=V20CredFormat.Format.INDY)
                ],
                attachments=[AttachDecorator.data_base64(TEST_INDY_FILTER, ident="indy")],
            )
            == TEST_INDY_FILTER
        )

        assert (
            V20CredFormat.Format.INDY.get_attachment_data(
                formats=[
                    V20CredFormat(attach_id="indy", format_=V20CredFormat.Format.INDY)
                ],
                attachments=[AttachDecorator.data_base64(TEST_INDY_FILTER, ident="xxx")],
            )
            is None
        )

        assert (
            V20CredFormat.Format.LD_PROOF.get_attachment_data(
                formats=[
                    V20CredFormat(attach_id="indy", format_=V20CredFormat.Format.INDY)
                ],
                attachments=[AttachDecorator.data_base64(TEST_INDY_FILTER, ident="indy")],
            )
            is None
        )



from datetime import datetime, timezone
from unittest import TestCase

from marshmallow.utils import INCLUDE

from ....ld_proofs.constants import CREDENTIALS_CONTEXT_V1_URL, VERIFIABLE_CREDENTIAL_TYPE
from ...models.credential import VerifiableCredential
from ...models.linked_data_proof import LDProof

CREDENTIAL = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://w3id.org/citizenship/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "id": "https://issuer.oidp.uscis.gov/credentials/83627465",
    "type": ["VerifiableCredential", "PermanentResidentCard"],
    "issuer": "did:example:489398593",
    "identifier": "83627465",
    "name": "Permanent Resident Card",
    "description": "Government of Example Permanent Resident Card.",
    "issuanceDate": "2019-12-03T12:19:52Z",
    "expirationDate": "2029-12-03T12:19:52Z",
    "credentialSubject": {
        "id": "did:example:b34ca6cd37bbf23",
        "type": ["PermanentResident", "Person"],
        "givenName": "JOHN",
        "familyName": "SMITH",
        "gender": "Male",
        "image": "data:image/png;base64,iVBORw0KGgokJggg==",
        "residentSince": "2015-01-01",
        "lprCategory": "C09",
        "lprNumber": "999-999-999",
        "commuterClassification": "C1",
        "birthCountry": "Bahamas",
        "birthDate": "1958-07-17",
    },
}


VC_PROOF = {
    "type": "BbsBlsSignatureProof2020",
    "created": "2020-12-14T03:11:30Z",
    "nonce": "CF69iO3nfvqRsRBNElE8b4wO39SyJHPM7Gg1nExltW5vSfQA1lvDCR/zXX1To0/4NLo=",
    "proofPurpose": "assertionMethod",
    "proofValue": "AA0f/7dHROgM2jXZiK0UmY5956/26qbbWF1sKgTMLx1NWEJqrE2ptwlREsxxrqZDRy5pxIFeSxDe08yWxDIk7zefzbwHd04hfbs0oaE2e9TMxIhfUZnct5Br7XenOwpZkkW1d7nt/yUFclgLCAIg+8B3UDpsuzv4rAJ3bTvD69nrMJPwC+Ao7meBgPcAaubNirSqrgAAAHSqzxvoLIRtX8mcq90yIHHuAcThiP63ChKE9c49pJboQ5FBA1aiMIIAJ+J7JPZtBGUAAAACIly7gNiA2nXJAVTKNepEQOtdyEU1gqExcaxWhMgX6nBCRGCwypy5lDDj2XWsvcuzPcvrpvaBxIBvTBAVjKDODaExOe1FKwA2t6F80wvt1BrEQpa5mG9YsI7Hw0wwl+c0SekC/WYlVW0oFjdICH+ZsAAAAAJlYX1Br69N/IAemIkmBvU/7bcIGssDcGL4hNzuTe0a8FnXYhUHyYmnMYFgZMv2ht2nMZiSwAugP2y3dFAU99bU",
    "verificationMethod": "did:example:489398593#test",
}

VERIFIABLE_CREDENTIAL = {**CREDENTIAL, "proof": VC_PROOF}


class TestLinkedDataProof(TestCase):
    """LinkedDataProof tests"""

    def test_serde(self):
        """Test de/serialization."""
        proof = LDProof.deserialize(VC_PROOF)
        assert type(proof) is LDProof

        proof_dict = proof.serialize()
        assert proof_dict == VC_PROOF


class TestVerifiableCredential(TestCase):
    """VerifiableCredential tests"""

    def test_serde_credential(self):
        """Test de/serialization."""
        credential = VerifiableCredential.deserialize(CREDENTIAL, unknown=INCLUDE)
        assert type(credential) is VerifiableCredential

        credential_dict = credential.serialize()
        assert credential_dict == CREDENTIAL

    def test_serde_verifiable_credential(self):
        """Test de/serialization."""
        credential = VerifiableCredential.deserialize(
            VERIFIABLE_CREDENTIAL, unknown=INCLUDE
        )
        assert type(credential) is VerifiableCredential

        credential_dict = credential.serialize()
        assert credential_dict == VERIFIABLE_CREDENTIAL

    def test_properties(self):
        credential = VerifiableCredential()

        credential.context = [CREDENTIALS_CONTEXT_V1_URL]
        assert credential.context == [CREDENTIALS_CONTEXT_V1_URL]

        credential.add_context("https://some.context")
        credential.add_context({"some": "context"})
        assert credential.context == [
            CREDENTIALS_CONTEXT_V1_URL,
            "https://some.context",
            {"some": "context"},
        ]

        assert credential.context_urls == [
            CREDENTIALS_CONTEXT_V1_URL,
            "https://some.context",
        ]

        with self.assertRaises(Exception):
            credential.context = ["IsNotVCContext"]

        credential.type = [VERIFIABLE_CREDENTIAL_TYPE]
        assert credential.type == [VERIFIABLE_CREDENTIAL_TYPE]

        credential.add_type("Sometype")
        assert credential.type == [VERIFIABLE_CREDENTIAL_TYPE, "Sometype"]

        with self.assertRaises(Exception):
            credential.type = ["DoesNotIncludeCredentialType"]

        credential.id = "http://someid.com"
        assert credential.id == "http://someid.com"

        credential.id = None
        assert not credential.id

        with self.assertRaises(Exception):
            credential.id = "not-an-uri"

        assert not credential.issuer_id
        credential.issuer_id = "http://some_id.com"
        assert credential.issuer_id == "http://some_id.com"

        assert credential.issuer == "http://some_id.com"
        credential.issuer = {"id": "http://some_id.com"}
        assert credential.issuer == {"id": "http://some_id.com"}
        assert credential.issuer_id == "http://some_id.com"
        credential.issuer_id = "http://some-other-id"
        assert credential.issuer == {"id": "http://some-other-id"}

        with self.assertRaises(Exception):
            credential.issuer = {"id": "not-an-uri"}

        with self.assertRaises(Exception):
            credential.issuer = "not-an-uri"

        with self.assertRaises(Exception):
            credential.issuer_id = "not-an-uri"

        with self.assertRaises(Exception):
            credential.issuer = {"not-id": "not-id"}

        date = datetime.now(timezone.utc)
        credential.issuance_date = date
        assert credential.issuance_date == date.isoformat()
        credential.issuance_date = date.isoformat()
        assert credential.issuance_date == date.isoformat()
        date = datetime(2019, 12, 11, 3, 50, 55, 0)
        credential.issuance_date = date
        assert (
            credential.issuance_date
            == datetime(2019, 12, 11, 3, 50, 55, 0, timezone.utc).isoformat()
        )

        date = datetime.now(timezone.utc)
        credential.expiration_date = date
        assert credential.expiration_date == date.isoformat()
        credential.expiration_date = date.isoformat()
        assert credential.expiration_date == date.isoformat()
        date = datetime(2019, 12, 11, 3, 50, 55, 0)
        credential.expiration_date = date
        assert (
            credential.expiration_date
            == datetime(2019, 12, 11, 3, 50, 55, 0, timezone.utc).isoformat()
        )

        assert not credential.credential_subject
        assert not credential.credential_subject_ids
        credential.credential_subject = {"some": "props"}
        assert credential.credential_subject == {"some": "props"}
        assert credential.credential_subject_ids == []
        credential.credential_subject = [{"some": "props"}]
        assert credential.credential_subject == [{"some": "props"}]
        assert credential.credential_subject_ids == []

        credential.credential_subject = {"id": "some:uri"}
        assert credential.credential_subject_ids == ["some:uri"]

        assert credential == credential
        assert credential != 10



"""Linked data proof signing and verification methods."""

from typing import List, Optional

from .document_loader import DocumentLoaderMethod
from .proof_set import ProofSet
from .purposes import _ProofPurpose as ProofPurpose
from .suites import _LinkedDataProof as LinkedDataProof
from .validation_result import DocumentVerificationResult


async def sign(
    *,
    document: dict,
    suite: LinkedDataProof,
    purpose: ProofPurpose,
    document_loader: DocumentLoaderMethod,
) -> dict:
    """Cryptographically signs the provided document by adding a `proof` section.

    Proof is added based on the provided suite and proof purpose

    Args:
        document (dict): JSON-LD document to be signed.
        suite (LinkedDataProof): The linked data signature cryptographic suite
            with which to sign the document
        purpose (ProofPurpose): A proof purpose instance that will match proofs to be
            verified and ensure they were created according to the appropriate purpose.
        document_loader (DocumentLoader): The document loader to use.

    Raises:
        LinkedDataProofException: When a jsonld url cannot be resolved, OR signing fails.

    Returns:
        dict: Signed document.

    """
    return await ProofSet.add(
        document=document,
        suite=suite,
        purpose=purpose,
        document_loader=document_loader,
    )


async def verify(
    *,
    document: dict,
    suites: List[LinkedDataProof],
    purpose: ProofPurpose,
    document_loader: DocumentLoaderMethod,
) -> DocumentVerificationResult:
    """Verify the linked data signature on the provided document.

    Args:
        document (dict): The document with one or more proofs to be verified.
        suites (List[LinkedDataProof]): Acceptable signature suite instances for
            verifying the proof(s).
        purpose (ProofPurpose): A proof purpose instance that will match proofs to be
            verified and ensure they were created according to the appropriate purpose.
        document_loader (DocumentLoader): The document loader to use.

    Returns:
        DocumentVerificationResult: Object with a `verified` boolean property that is
            `True` if at least one proof matching the given purpose and suite verifies
            and `False` otherwise. a `results` property with an array of detailed
            results. if `False` an `errors` property will be present, with a list
            containing all of the errors that occurred during the verification process.

    """

    result = await ProofSet.verify(
        document=document,
        suites=suites,
        purpose=purpose,
        document_loader=document_loader,
    )

    return result


async def derive(
    *,
    document: dict,
    reveal_document: dict,
    suite: LinkedDataProof,
    document_loader: DocumentLoaderMethod,
    nonce: Optional[bytes] = None,
) -> dict:
    """Derive proof(s) for document with reveal document.

    All proofs matching the signature suite type will be replaced with a derived
    proof. Other proofs will be discarded.

    Args:
        document (dict): The document with one or more proofs to be derived
        reveal_document (dict): The JSON-LD frame specifying the revealed attributes
        suite (LinkedDataProof): The linked data signature cryptographic suite
            with which to derive the proof
        document_loader (DocumentLoader): The document loader to use.
        nonce (bytes, optional): Nonce to use for the proof. Defaults to None.

    Returns:
        dict: The document with derived proof(s).

    """

    result = await ProofSet.derive(
        document=document,
        reveal_document=reveal_document,
        suite=suite,
        document_loader=document_loader,
        nonce=nonce,
    )

    return result



"""Cred request artifacts to attach to RFC 453 messages."""

from typing import Mapping, Optional

from marshmallow import EXCLUDE, fields

from ...messaging.models.base import BaseModel, BaseModelSchema
from ...messaging.valid import (
    INDY_CRED_DEF_ID_EXAMPLE,
    INDY_CRED_DEF_ID_VALIDATE,
    NUM_STR_WHOLE_EXAMPLE,
    NUM_STR_WHOLE_VALIDATE,
    UUID4_EXAMPLE,
)


class IndyCredRequest(BaseModel):
    """Indy credential request."""

    class Meta:
        """Indy credential request metadata."""

        schema_class = "IndyCredRequestSchema"

    def __init__(
        self,
        prover_did: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        blinded_ms: Optional[Mapping] = None,
        blinded_ms_correctness_proof: Optional[Mapping] = None,
        nonce: Optional[str] = None,
        **kwargs,
    ):
        """Initialize indy credential request."""
        super().__init__(**kwargs)
        self.prover_did = prover_did
        self.cred_def_id = cred_def_id
        self.blinded_ms = blinded_ms
        self.blinded_ms_correctness_proof = blinded_ms_correctness_proof
        self.nonce = nonce


class IndyCredRequestSchema(BaseModelSchema):
    """Indy credential request schema."""

    class Meta:
        """Indy credential request schema metadata."""

        model_class = IndyCredRequest
        unknown = EXCLUDE

    prover_did = fields.Str(
        required=True,
        metadata={
            "description": "Prover DID/Random String/UUID",
            "example": UUID4_EXAMPLE,
        },
    )
    cred_def_id = fields.Str(
        required=True,
        validate=INDY_CRED_DEF_ID_VALIDATE,
        metadata={
            "description": "Credential definition identifier",
            "example": INDY_CRED_DEF_ID_EXAMPLE,
        },
    )
    blinded_ms = fields.Dict(
        required=True, metadata={"description": "Blinded master secret"}
    )
    blinded_ms_correctness_proof = fields.Dict(
        required=True,
        metadata={"description": "Blinded master secret correctness proof"},
    )
    nonce = fields.Str(
        required=True,
        validate=NUM_STR_WHOLE_VALIDATE,
        metadata={
            "description": "Nonce in credential request",
            "example": NUM_STR_WHOLE_EXAMPLE,
        },
    )



"""Classes for tracking performance and timing."""

import functools
import inspect
import time
from typing import Optional, Sequence, TextIO, Union


class Stats:
    """A collection of statistics."""

    def __init__(self):
        """Initialize the Stats instance."""
        self.counts = {}
        self.max_time = {}
        self.min_time = {}
        self.total_time = {}

    def log(self, name: str, duration: float):
        """Log an entry in the stats."""
        if name in self.counts:
            self.counts[name] += 1
            self.max_time[name] = max(self.max_time[name], duration)
            self.min_time[name] = min(self.min_time[name], duration)
            self.total_time[name] += duration
        else:
            self.counts[name] = 1
            self.max_time[name] = duration
            self.min_time[name] = duration
            self.total_time[name] = duration

    def extract(self, names: Sequence[str] = None) -> dict:
        """Summarize the stats in a dictionary."""
        counts = self.counts.copy()
        all_names = set(counts)
        if names is None:
            names = all_names
            maxes = self.max_time.copy()
            mins = self.min_time.copy()
            totals = self.total_time.copy()
        else:
            names = set(names).intersection(all_names)
            counts = {name: val for (name, val) in counts.items() if name in names}
            maxes = {name: val for (name, val) in self.max_time.items() if name in names}
            mins = {name: val for (name, val) in self.min_time.items() if name in names}
            totals = {
                name: val for (name, val) in self.total_time.items() if name in names
            }

        return {
            "avg": {name: totals[name] / counts[name] for name in names},
            "count": counts,
            "max": maxes,
            "min": mins,
            "total": totals,
        }


class Timer:
    """Timer instance for a running task."""

    def __init__(self, collector: "Collector", groups: Sequence[str]):
        """Initialize the Timer instance."""
        self.collector = collector
        self.groups = groups
        self.start_time = None

    @classmethod
    def now(cls):
        """Fetch a standard timer value."""
        return time.perf_counter()

    def start(self) -> "Timer":
        """Start the timer."""
        self.start_time = self.now()
        return self

    def stop(self):
        """Stop the timer."""
        if self.start_time:
            dur = self.now() - self.start_time
            for grp in self.groups:
                self.collector.log(grp, dur, self.start_time)
        self.start_time = None

    def __enter__(self):
        """Enter the context manager."""
        return self.start()

    def __exit__(self, type, value, tb):
        """Exit the context manager."""
        self.stop()


class Collector:
    """Collector for a set of statistics."""

    def __init__(self, *, enabled: bool = True, log_path: Optional[str] = None):
        """Initialize the Collector instance."""
        self._enabled = enabled
        self._log_file: Optional[TextIO] = None
        self._log_path = log_path
        self._stats = None
        self.reset()

    def reset(self):
        """Reset the collector's statistics."""
        self._stats = Stats()
        if self._log_file:
            self._log_file.close()
            self._log_file = None
        if self._log_path:
            self._log_file = open(self._log_path, "w")

    @property
    def enabled(self) -> bool:
        """Accessor for the collector's enabled property."""
        return self._enabled

    @enabled.setter
    def enabled(self, val: bool):
        """Setter for the collector's enabled property."""
        self._enabled = val

    def log(self, name: str, duration: float, start: Optional[float] = None):
        """Log an entry in the statistics if the collector is enabled."""
        if self._enabled:
            self._stats.log(name, duration)
            if self._log_file:
                if start is None:
                    start = time.perf_counter() - duration
                self._log_file.write(f"{name} {start:.5f} {duration:.5f}\n")

    def mark(self, *names):
        """Make a custom decorator function for adding to the set of groups."""
        return lambda fn: self(fn, names)

    def wrap(
        self,
        obj,
        prop_name: Union[str, Sequence[str]],
        groups: Sequence[str] = None,
        *,
        ignore_missing: bool = False,
    ):
        """Wrap a method on a class or class instance."""
        if not prop_name:
            raise ValueError("missing prop_name")
        if isinstance(prop_name, str):
            method = getattr(obj, prop_name, None)
            if method:
                setattr(obj, prop_name, self(method, groups))
            elif not ignore_missing:
                raise AttributeError(prop_name)
        else:
            for prop in prop_name:
                self.wrap(obj, prop, groups)

    def wrap_fn(self, fn, groups: Sequence[str]):
        """Wrap a function instance to collect timing statistics on execution."""

        @functools.wraps(fn)
        def wrapped(*args, **kwargs):
            with self.timer(*groups):
                result = fn(*args, **kwargs)
            return result

        return wrapped

    def wrap_coro(self, fn, groups: Sequence[str]):
        """Wrap a coroutine instance to collect timing statistics on execution."""

        @functools.wraps(fn)
        async def wrapped(*args, **kwargs):
            with self.timer(*groups):
                result = await fn(*args, **kwargs)
            return result

        return wrapped

    def __call__(self, fn, groups: Sequence[str] = None):
        """Decorate a function or class method.

        Returns: a wrapped function or coroutine with automatic stats collection
        """
        groups = set(groups) if groups else set()
        if inspect.iscoroutinefunction(fn):
            groups.add(fn.__qualname__)
            return self.wrap_coro(fn, groups)
        elif inspect.isfunction(fn) or inspect.ismethod(fn):
            groups.add(fn.__qualname__)
            return self.wrap_fn(fn, groups)
        else:
            raise ValueError(f"Expected function or coroutine, got: {fn}")

    def timer(self, *groups):
        """Create a new timer attached to this collector."""
        return Timer(self, groups)

    @property
    def results(self) -> dict:
        """Accessor for the current set of collected statistics."""
        return self._stats.extract()

    def extract(self, groups: Sequence[str] = None) -> dict:
        """Extract statistics for a specific set of groups."""
        return self._stats.extract(groups)



"""Test keylist query message."""

from unittest import TestCase

from ...message_types import KEYLIST_QUERY
from ..inner.keylist_query_paginate import KeylistQueryPaginate
from ..keylist_query import KeylistQuery, KeylistQuerySchema
from . import MessageTest


class TestKeylistQuery(MessageTest, TestCase):
    """Test KeylistQuery message."""

    TYPE = KEYLIST_QUERY
    CLASS = KeylistQuery
    SCHEMA = KeylistQuerySchema
    VALUES = {
        "filter": {"filter": "something"},
        "paginate": KeylistQueryPaginate(10, 10),
    }



from io import BufferedReader, TextIOWrapper
from tempfile import NamedTemporaryFile
from unittest import IsolatedAsyncioTestCase, mock

from ..logging import configurator as test_module


class TestLoggingConfigurator(IsolatedAsyncioTestCase):
    agent_label_arg_value = "Aries Cloud Agent"
    transport_arg_value = "transport"
    host_arg_value = "host"
    port_arg_value = "port"

    @mock.patch.object(test_module, "load_resource", autospec=True)
    @mock.patch.object(test_module, "fileConfig", autospec=True)
    def test_configure_default(self, mock_file_config, mock_load_resource):
        test_module.LoggingConfigurator.configure()

        mock_load_resource.assert_called_once_with(
            test_module.DEFAULT_LOGGING_CONFIG_PATH_INI, "utf-8"
        )
        mock_file_config.assert_called_once_with(
            mock_load_resource.return_value,
            disable_existing_loggers=False,
            new_file_path=None,
        )

    def test_configure_default_with_log_file_error_level(self):
        log_file = NamedTemporaryFile()
        with mock.patch.object(
            test_module, "load_resource", mock.MagicMock()
        ) as mock_load:
            mock_load.return_value = None
            test_module.LoggingConfigurator.configure(
                log_level="ERROR", log_file=log_file.name
            )

    @mock.patch.object(test_module, "load_resource", autospec=True)
    @mock.patch.object(test_module, "fileConfig", autospec=True)
    def test_configure_default_with_path(self, mock_file_config, mock_load_resource):
        path = "a path"
        test_module.LoggingConfigurator.configure(path)

        mock_load_resource.assert_called_once_with(path, "utf-8")
        mock_file_config.assert_called_once_with(
            mock_load_resource.return_value,
            disable_existing_loggers=False,
            new_file_path=None,
        )

    def test_configure_multitenant(self):
        with mock.patch.object(
            test_module,
            "logging",
            mock.MagicMock(
                basicConfig=mock.MagicMock(),
                FileHandler=mock.MagicMock(),
                root=mock.MagicMock(
                    warning=mock.MagicMock(),
                    handlers=[],
                ),
            ),
        ):
            test_module.LoggingConfigurator.configure(
                log_file="test.log",
                multitenant=True,
            )

    def test_configure_with_multitenant_with_yaml_file(self):
        with mock.patch.object(
            test_module,
            "logging",
            mock.MagicMock(
                basicConfig=mock.MagicMock(),
                FileHandler=mock.MagicMock(),
                root=mock.MagicMock(
                    warning=mock.MagicMock(),
                    handlers=[],
                ),
            ),
        ):
            test_module.LoggingConfigurator.configure(
                log_config_path="acapy_agent/config/logging/default_multitenant_logging_config.yml",
                log_file="test.log",
                multitenant=True,
            )

    def test_banner_did(self):
        mock_http = mock.MagicMock(scheme="http", host="1.2.3.4", port=8081)
        mock_https = mock.MagicMock(schemes=["https", "archie"])
        mock_admin_server = mock.MagicMock(host="1.2.3.4", port=8091)
        test_label = "Aries Cloud Agent"
        test_did = "55GkHamhTU1ZbTbV2ab9DE"

        with self.assertLogs(level="INFO") as log:
            test_module.LoggingConfigurator.print_banner(
                test_label,
                {"in": mock_http},
                {"out": mock_https},
                test_did,
                mock_admin_server,
            )
            test_module.LoggingConfigurator.print_banner(
                test_label, {"in": mock_http}, {"out": mock_https}, test_did
            )

        # Join all log records and check if DID is present
        output = "\n".join(log.output)
        assert test_did in output

    def test_load_resource(self):
        # Testing local file access
        with mock.patch("builtins.open", mock.MagicMock()) as mock_open:
            # First call succeeds
            file_handle = mock.MagicMock(spec=TextIOWrapper)
            mock_open.return_value = file_handle
            result = test_module.load_resource("abc", encoding="utf-8")
            mock_open.assert_called_once_with("abc", encoding="utf-8")
            assert result == file_handle  # Verify the returned file handle

            mock_open.reset_mock()
            # Simulate IOError on second call
            mock_open.side_effect = IOError("insufficient privilege")
            # load_resource should absorb IOError
            result = test_module.load_resource("abc", encoding="utf-8")
            mock_open.assert_called_once_with("abc", encoding="utf-8")
            assert result is None

        # Testing package resource access with encoding (text mode)
        with (
            mock.patch("importlib.resources.files") as mock_files,
            mock.patch("io.TextIOWrapper", mock.MagicMock()) as mock_text_io_wrapper,
        ):
            # Setup the mocks
            mock_resource_path = mock.MagicMock()
            mock_files.return_value.joinpath.return_value = mock_resource_path
            mock_resource_handle = mock.MagicMock(spec=BufferedReader)
            mock_resource_path.open.return_value = mock_resource_handle
            mock_text_io_wrapper.return_value = mock.MagicMock(spec=TextIOWrapper)

            result = test_module.load_resource("abc:def", encoding="utf-8")

            # Assertions
            mock_files.assert_called_once_with("abc")
            mock_files.return_value.joinpath.assert_called_once_with("def")
            mock_resource_path.open.assert_called_once_with("rb")
            mock_text_io_wrapper.assert_called_once_with(
                mock_resource_handle, encoding="utf-8"
            )
            assert result is mock_text_io_wrapper.return_value

        # Testing package resource access without encoding (binary mode)
        with mock.patch("importlib.resources.files") as mock_files:
            # Setup the mocks
            mock_resource_path = mock.MagicMock()
            mock_files.return_value.joinpath.return_value = mock_resource_path
            mock_resource_handle = mock.MagicMock(spec=BufferedReader)
            mock_resource_path.open.return_value = mock_resource_handle

            result = test_module.load_resource("abc:def", encoding=None)

            # Assertions
            mock_files.assert_called_once_with("abc")
            mock_files.return_value.joinpath.assert_called_once_with("def")
            mock_resource_path.open.assert_called_once_with("rb")
            assert result == mock_resource_handle  # Verify the returned binary stream



"""Credential format inner object."""

from collections import namedtuple
from enum import Enum
from typing import TYPE_CHECKING, Mapping, Optional, Sequence, Type, Union

from marshmallow import EXCLUDE, fields
from uuid_utils import uuid4

from .....messaging.decorators.attach_decorator import AttachDecorator
from .....messaging.models.base import BaseModel, BaseModelSchema
from .....messaging.valid import UUID4_EXAMPLE
from .....utils.classloader import DeferLoad

if TYPE_CHECKING:
    from ..formats.handler import V20PresFormatHandler

# aries prefix
FormatSpec = namedtuple("FormatSpec", "aries handler")


class V20PresFormat(BaseModel):
    """Present-proof protocol message attachment format."""

    class Meta:
        """Present-proof protocol message attachment format metadata."""

        schema_class = "V20PresFormatSchema"

    class Format(Enum):
        """Attachment format."""

        ANONCREDS = FormatSpec(
            "anoncreds/",
            DeferLoad(
                "acapy_agent.protocols.present_proof.v2_0"
                ".formats.anoncreds.handler.AnonCredsPresExchangeHandler"
            ),
        )
        INDY = FormatSpec(
            "hlindy/",
            DeferLoad(
                "acapy_agent.protocols.present_proof.v2_0"
                ".formats.indy.handler.IndyPresExchangeHandler"
            ),
        )
        DIF = FormatSpec(
            "dif/",
            DeferLoad(
                "acapy_agent.protocols.present_proof.v2_0"
                ".formats.dif.handler.DIFPresFormatHandler"
            ),
        )

        @classmethod
        def get(cls, label: Union[str, "V20PresFormat.Format"]):
            """Get format enum for label."""
            if isinstance(label, str):
                for fmt in V20PresFormat.Format:
                    if label.startswith(fmt.aries) or label == fmt.api:
                        return fmt
            elif isinstance(label, V20PresFormat.Format):
                return label

            return None

        @property
        def api(self) -> str:
            """Admin API specifier."""
            return self.name.lower()

        @property
        def aries(self) -> str:
            """Accessor for aries identifier."""
            return self.value.aries

        @property
        def handler(self) -> Type["V20PresFormatHandler"]:
            """Accessor for presentation exchange format handler."""
            return self.value.handler.resolved

        def validate_fields(self, message_type: str, attachment_data: Mapping):
            """Raise ValidationError for invalid attachment formats."""
            self.handler.validate_fields(message_type, attachment_data)

        def get_attachment_data(
            self,
            formats: Sequence["V20PresFormat"],
            attachments: Sequence[AttachDecorator],
        ):
            """Find attachment of current format, decode and return its content."""
            for fmt in formats:
                if V20PresFormat.Format.get(fmt.format) is self:
                    attach_id = fmt.attach_id
                    break
            else:
                return None

            for atch in attachments:
                if atch.ident == attach_id:
                    return atch.content

            return None

    def __init__(
        self,
        *,
        attach_id: Optional[str] = None,
        format_: Optional[str] = None,
    ):
        """Initialize present-proof protocol message attachment format."""
        self.attach_id = attach_id or uuid4()
        self.format_ = format_

    @property
    def format(self) -> str:
        """Return format."""
        return self.format_


class V20PresFormatSchema(BaseModelSchema):
    """Present-proof protocol message attachment format schema."""

    class Meta:
        """Present-proof protocol message attachment format schema metadata."""

        model_class = V20PresFormat
        unknown = EXCLUDE

    attach_id = fields.Str(
        required=True,
        allow_none=False,
        metadata={"description": "Attachment identifier", "example": UUID4_EXAMPLE},
    )
    format_ = fields.Str(
        required=True,
        allow_none=False,
        data_key="format",
        metadata={
            "description": "Attachment format specifier",
            "example": "dif/presentation-exchange/submission@v1.0",
        },
    )



"""Admin request context class.

A request context provided by the admin server to admin route handlers.
"""

from typing import Mapping, Optional, Type

from ..config.injection_context import InjectionContext
from ..config.injector import InjectionError, Injector, InjectType
from ..config.settings import Settings
from ..core.profile import Profile, ProfileSession


class AdminRequestContext:
    """Context established by the Conductor and passed into message handlers."""

    def __init__(
        self,
        profile: Profile,
        *,
        context: Optional[InjectionContext] = None,
        settings: Optional[Mapping[str, object]] = None,
        root_profile: Optional[Profile] = None,
        metadata: Optional[dict] = None,
    ):
        """Initialize an instance of AdminRequestContext."""
        self._context = (context or profile.context).start_scope(settings)
        self._profile = profile
        self._root_profile = root_profile
        self._metadata = metadata

    @property
    def injector(self) -> Injector:
        """Accessor for the associated `Injector` instance."""
        return self._context.injector

    @property
    def profile(self) -> Profile:
        """Accessor for the associated `Profile` instance."""
        return self._profile

    @property
    def root_profile(self) -> Optional[Profile]:
        """Accessor for the associated root_profile instance."""
        return self._root_profile

    @property
    def metadata(self) -> dict:
        """Accessor for the associated metadata."""
        return self._metadata

    @property
    def settings(self) -> Settings:
        """Accessor for the context settings."""
        return self._context.settings

    def session(self) -> ProfileSession:
        """Start a new interactive session with no transaction support requested."""
        return self.profile.session(self._context)

    def transaction(self) -> ProfileSession:
        """Start a new interactive session with commit and rollback support.

        If the current backend does not support transactions, then commit
        and rollback operations of the session will not have any effect.
        """
        return self.profile.transaction(self._context)

    def inject(
        self,
        base_cls: Type[InjectType],
        settings: Optional[Mapping[str, object]] = None,
    ) -> InjectType:
        """Get the provided instance of a given class identifier.

        Args:
            cls: The base class to retrieve an instance of
            base_cls: The base class to retrieve an instance of
            settings: An optional mapping providing configuration to the provider

        Returns:
            An instance of the base class, or None

        """
        return self._context.inject(base_cls, settings)

    def inject_or(
        self,
        base_cls: Type[InjectType],
        settings: Optional[Mapping[str, object]] = None,
        default: Optional[InjectType] = None,
    ) -> Optional[InjectType]:
        """Get the provided instance of a given class identifier or default if not found.

        Args:
            base_cls: The base class to retrieve an instance of
            settings: An optional dict providing configuration to the provider
            default: default return value if no instance is found

        Returns:
            An instance of the base class, or None

        """
        return self._context.inject_or(base_cls, settings, default)

    def update_settings(self, settings: Mapping[str, object]):
        """Update the current scope with additional settings."""
        self._context.update_settings(settings)

    @classmethod
    def test_context(
        cls, session_inject: dict, profile: Profile
    ) -> "AdminRequestContext":
        """Quickly set up a new admin request context for tests."""
        ctx = AdminRequestContext(profile)
        setattr(ctx, "session_inject", {} if session_inject is None else session_inject)
        setattr(ctx, "session", ctx._test_session)
        return ctx

    def _test_session(self) -> ProfileSession:
        session = self.profile.session(self._context)

        def _inject(base_cls):
            if session._active and base_cls in self.session_inject:
                ret = self.session_inject[base_cls]
                if ret is None:
                    raise InjectionError(
                        "No instance provided for class: {}".format(base_cls.__name__)
                    )
                return ret
            return session._context.injector.inject(base_cls)

        def _inject_or(base_cls, default=None):
            if session._active and base_cls in self.session_inject:
                ret = self.session_inject[base_cls]
                if ret is None:
                    ret = default
                return ret
            return session._context.injector.inject_or(base_cls, default)

        setattr(session, "inject", _inject)
        setattr(session, "inject_or", _inject_or)
        return session

    def __repr__(self) -> str:
        """Provide a human readable representation of this object.

        Returns:
            A human readable representation of this object

        """
        skip = ("session",)
        items = (
            "{}={}".format(k, repr(v)) for k, v in self.__dict__.items() if k not in skip
        )
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))



from unittest import TestCase

from ..base import Role


class TestRole(TestCase):
    def test_role(self):
        assert Role.get(2) is Role.STEWARD
        assert Role.get(0) is Role.TRUSTEE
        assert Role.get(101) is Role.ENDORSER
        assert Role.get(201) is Role.NETWORK_MONITOR
        assert Role.get(None) is Role.USER
        assert Role.get(-1) is None
        assert Role.get("user") is Role.USER
        assert Role.get("steward") is Role.STEWARD
        assert Role.get("trustee") is Role.TRUSTEE
        assert Role.get("endorser") is Role.ENDORSER
        assert Role.get("network_monitor") is Role.NETWORK_MONITOR
        assert Role.get("ROLE_REMOVE") is None

        assert Role.STEWARD.to_indy_num_str() == "2"
        assert Role.TRUSTEE.to_indy_num_str() == "0"
        assert Role.ENDORSER.to_indy_num_str() == "101"
        assert Role.NETWORK_MONITOR.to_indy_num_str() == "201"
        assert Role.USER.to_indy_num_str() is None
        assert Role.ROLE_REMOVE.to_indy_num_str() == ""

        assert Role.STEWARD.token() == "STEWARD"
        assert Role.TRUSTEE.token() == "TRUSTEE"
        assert Role.ENDORSER.token() == "ENDORSER"
        assert Role.NETWORK_MONITOR.token() == "NETWORK_MONITOR"
        assert Role.USER.token() is None
        assert Role.ROLE_REMOVE.to_indy_num_str() == ""



from unittest import IsolatedAsyncioTestCase

from ....storage.base import StorageNotFoundError
from ....utils.testing import create_test_profile
from .. import issuer_cred_rev_record as test_module
from ..issuer_cred_rev_record import IssuerCredRevRecord

TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
CRED_DEF_ID = f"{TEST_DID}:3:CL:1234:default"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:0"


class TestIssuerCredRevRecord(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

    async def test_serde(self):
        rec = IssuerCredRevRecord(
            record_id=test_module.UUID4_EXAMPLE,
            state=IssuerCredRevRecord.STATE_ISSUED,
            cred_ex_id=test_module.UUID4_EXAMPLE,
            rev_reg_id=REV_REG_ID,
            cred_rev_id="1",
        )
        ser = rec.serialize()
        assert ser["record_id"] == rec.record_id
        assert ser["cred_def_id"] == CRED_DEF_ID
        assert rec.cred_def_id == CRED_DEF_ID

        assert rec == IssuerCredRevRecord.deserialize(ser)

    async def test_rec_ops(self):
        recs = [
            IssuerCredRevRecord(
                state=IssuerCredRevRecord.STATE_ISSUED,
                cred_ex_id=test_module.UUID4_EXAMPLE,
                rev_reg_id=REV_REG_ID,
                cred_rev_id=str(i + 1),
            )
            for i in range(2)
        ]
        async with self.profile.session() as session:
            await recs[0].set_state(
                session,
                IssuerCredRevRecord.STATE_REVOKED,
            )  # saves
            assert recs[0] != recs[1]

            assert (await IssuerCredRevRecord.query_by_ids(session))[0] == recs[0]
            assert (
                await IssuerCredRevRecord.retrieve_by_cred_ex_id(
                    session, test_module.UUID4_EXAMPLE
                )
            ) == recs[0]
            assert (
                await IssuerCredRevRecord.query_by_ids(
                    session,
                    cred_def_id=CRED_DEF_ID,
                )
            )[0] == recs[0]
            assert (
                await IssuerCredRevRecord.query_by_ids(
                    session,
                    rev_reg_id=REV_REG_ID,
                )
            )[0] == recs[0]
            assert (
                await IssuerCredRevRecord.query_by_ids(
                    session,
                    cred_def_id=CRED_DEF_ID,
                    rev_reg_id=REV_REG_ID,
                )
            )[0] == recs[0]
            assert (
                await IssuerCredRevRecord.query_by_ids(
                    session,
                    state=IssuerCredRevRecord.STATE_REVOKED,
                )
            )[0] == recs[0]
            assert not (
                await IssuerCredRevRecord.query_by_ids(
                    session,
                    state=IssuerCredRevRecord.STATE_ISSUED,
                )
            )

            assert (
                await IssuerCredRevRecord.retrieve_by_ids(
                    session, rev_reg_id=REV_REG_ID, cred_rev_id="1"
                )
                == recs[0]
            )
            with self.assertRaises(StorageNotFoundError):
                await IssuerCredRevRecord.retrieve_by_ids(
                    session, rev_reg_id=REV_REG_ID, cred_rev_id="2"
                )



"""did:peer:1 resolver implementation."""

import logging
import re
from typing import Callable, Optional, Pattern, Sequence, Text, Union

from acapy_agent.messaging.valid import B58

from ...config.injection_context import InjectionContext
from ...connections.base_manager import BaseConnectionManager
from ...core.profile import Profile
from ...resolver.base import BaseDIDResolver, DIDNotFound, ResolverType
from ...storage.error import StorageNotFoundError

LOGGER = logging.getLogger(__name__)


# TODO Copy pasted from did-peer-4, reuse when available
def _operate_on_embedded(
    visitor: Callable[[dict], dict],
) -> Callable[[Union[dict, str]], Union[dict, str]]:
    """Return an adapter function that turns a vm visitor into a vm | ref visitor.

    The adapter function calls a visitor on embedded vms but just returns on refs.
    """

    def _adapter(vm: Union[dict, str]) -> Union[dict, str]:
        if isinstance(vm, dict):
            return visitor(vm)
        return vm

    return _adapter


def _visit_verification_methods(document: dict, visitor: Callable[[dict], dict]):
    """Visit all verification methods in a document.

    This includes the main verificationMethod list as well as verification
    methods embedded in relationships.
    """
    verification_methods = document.get("verificationMethod")
    if verification_methods:
        document["verificationMethod"] = [visitor(vm) for vm in verification_methods]

    for relationship in (
        "authentication",
        "assertionMethod",
        "keyAgreement",
        "capabilityInvocation",
        "capabilityDelegation",
    ):
        vms_and_refs = document.get(relationship)
        if vms_and_refs:
            document[relationship] = [
                _operate_on_embedded(visitor)(vm) for vm in vms_and_refs
            ]

    return document


def contextualize(did: str, document: dict):
    """Contextualize a peer DID document."""

    def _contextualize_verification_method(vm: dict):
        """Contextualize a verification method."""
        if vm["controller"] == "#id":
            vm["controller"] = did
        if vm["id"].startswith("#"):
            vm["id"] = f"{did}{vm['id']}"
        return vm

    document = _visit_verification_methods(document, _contextualize_verification_method)

    for service in document.get("service", []):
        if service["id"].startswith("#"):
            service["id"] = f"{did}{service['id']}"

    return document


class PeerDID1Resolver(BaseDIDResolver):
    """Resolve legacy peer DIDs."""

    PEER1_PATTERN = re.compile(rf"^did:peer:1zQm[{B58}]{{44}}$")

    def __init__(self):
        """Initialize the resolver instance."""
        super().__init__(ResolverType.NATIVE)

    async def setup(self, context: InjectionContext):
        """Perform required setup for the resolver."""

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported_did_regex of DID Peer 1 Resolver."""
        return self.PEER1_PATTERN

    async def _fetch_did_document(self, profile: Profile, did: str) -> Optional[dict]:
        """Fetch DID from wallet if available.

        This is the method to be used with fetch_did_document to enable caching.
        """
        conn_mgr = BaseConnectionManager(profile)
        try:
            doc, _ = await conn_mgr.fetch_did_document(did)
            LOGGER.debug("Fetched doc %s", doc)
            return doc
        except StorageNotFoundError:
            LOGGER.debug("Failed to fetch doc for did %s", did)

        return None

    async def _resolve(
        self,
        profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve Legacy Peer DID to a DID document by fetching from the wallet.

        By the time this resolver is selected, it should be impossible for it
        to raise a DIDNotFound.
        """
        result = await self._fetch_did_document(profile, did)
        if result:
            # Apply corrections?
            result = contextualize(did, result)
            LOGGER.debug("Resolved %s to %s", did, result)
            return result
        else:
            raise DIDNotFound(f"DID not found: {did}")



"""Test PeerDIDResolver."""

import pytest
import pytest_asyncio

from ....core.event_bus import EventBus
from ....core.profile import Profile
from ....utils.testing import create_test_profile
from .. import peer4 as test_module
from ..peer4 import PeerDID4Resolver

# https://identity.foundation/peer-did-method-spec/#method-4-short-form-and-long-form
TEST_LONG_DP4 = "did:peer:4zQmd8CpeFPci817KDsbSAKWcXAE2mjvCQSasRewvbSF54Bd:z2M1k7h4psgp4CmJcnQn2Ljp7Pz7ktsd7oBhMU3dWY5s4fhFNj17qcRTQ427C7QHNT6cQ7T3XfRh35Q2GhaNFZmWHVFq4vL7F8nm36PA9Y96DvdrUiRUaiCuXnBFrn1o7mxFZAx14JL4t8vUWpuDPwQuddVo1T8myRiVH7wdxuoYbsva5x6idEpCQydJdFjiHGCpNc2UtjzPQ8awSXkctGCnBmgkhrj5gto3D4i3EREXYq4Z8r2cWGBr2UzbSmnxW2BuYddFo9Yfm6mKjtJyLpF74ytqrF5xtf84MnGFg1hMBmh1xVx1JwjZ2BeMJs7mNS8DTZhKC7KH38EgqDtUZzfjhpjmmUfkXg2KFEA3EGbbVm1DPqQXayPYKAsYPS9AyKkcQ3fzWafLPP93UfNhtUPL8JW5pMcSV3P8v6j3vPXqnnGknNyBprD6YGUVtgLiAqDBDUF3LSxFQJCVYYtghMTv8WuSw9h1a1SRFrDQLGHE4UrkgoRvwaGWr64aM87T1eVGkP5Dt4L1AbboeK2ceLArPScrdYGTpi3BpTkLwZCdjdiFSfTy9okL1YNRARqUf2wm8DvkVGUU7u5nQA3ZMaXWJAewk6k1YUxKd7LvofGUK4YEDtoxN5vb6r1Q2godrGqaPkjfL3RoYPpDYymf9XhcgG8Kx3DZaA6cyTs24t45KxYAfeCw4wqUpCH9HbpD78TbEUr9PPAsJgXBvBj2VVsxnr7FKbK4KykGcg1W8M1JPz21Z4Y72LWgGQCmixovrkHktcTX1uNHjAvKBqVD5C7XmVfHgXCHj7djCh3vzLNuVLtEED8J1hhqsB1oCBGiuh3xXr7fZ9wUjJCQ1HYHqxLJKdYKtoCiPmgKM7etVftXkmTFETZmpM19aRyih3bao76LdpQtbw636r7a3qt8v4WfxsXJetSL8c7t24SqQBcAY89FBsbEnFNrQCMK3JEseKHVaU388ctvRD45uQfe5GndFxthj4iSDomk4uRFd1uRbywoP1tRuabHTDX42UxPjz"
TEST_SHORT_DP4 = "did:peer:4zQmd8CpeFPci817KDsbSAKWcXAE2mjvCQSasRewvbSF54Bd"


@pytest.fixture
def event_bus():
    yield EventBus()


@pytest_asyncio.fixture
async def profile(event_bus: EventBus):
    """Profile fixture."""
    profile = await create_test_profile()
    profile.context.injector.bind_instance(EventBus, event_bus)
    yield profile


@pytest_asyncio.fixture
async def resolver(profile):
    """Resolver fixture."""
    instance = PeerDID4Resolver()
    await instance.setup(profile.context)
    yield instance


@pytest.mark.asyncio
async def test_resolve_4(profile: Profile, resolver: PeerDID4Resolver):
    """Test resolver setup."""
    assert resolver.supported_did_regex
    assert await resolver.supports(profile, TEST_LONG_DP4)
    assert await resolver.supports(profile, TEST_SHORT_DP4)

    long_doc = await resolver.resolve(profile, TEST_LONG_DP4)
    short_doc = await resolver.resolve(profile, TEST_SHORT_DP4)

    assert long_doc["id"] == TEST_LONG_DP4
    assert TEST_SHORT_DP4 in long_doc["alsoKnownAs"]
    assert short_doc["id"] == TEST_SHORT_DP4


@pytest.mark.asyncio
async def test_resolve_x_no_long(profile: Profile, resolver: PeerDID4Resolver):
    """Test resolver setup."""
    with pytest.raises(test_module.DIDNotFound):
        await resolver.resolve(profile, TEST_SHORT_DP4)



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import PING
from ..ping import Ping


class TestPing(TestCase):
    def setUp(self):
        self.test_comment = "hello"
        self.test_response_requested = True
        self.test_ping = Ping(
            comment=self.test_comment, response_requested=self.test_response_requested
        )

    def test_init(self):
        """Test initialization."""
        assert self.test_ping.comment == self.test_comment
        assert self.test_ping.response_requested == self.test_response_requested

    def test_type(self):
        """Test type."""
        assert self.test_ping._type == DIDCommPrefix.qualify_current(PING)

    @mock.patch("acapy_agent.protocols.trustping.v1_0.messages.ping.PingSchema.load")
    def test_deserialize(self, mock_ping_schema_load):
        """
        Test deserialization.
        """
        obj = {"obj": "obj"}

        msg = Ping.deserialize(obj)
        mock_ping_schema_load.assert_called_once_with(obj)

        assert msg is mock_ping_schema_load.return_value

    @mock.patch("acapy_agent.protocols.trustping.v1_0.messages.ping.PingSchema.dump")
    def test_serialize(self, mock_ping_schema_load):
        """
        Test serialization.
        """
        msg_dict = self.test_ping.serialize()
        mock_ping_schema_load.assert_called_once_with(self.test_ping)

        assert msg_dict is mock_ping_schema_load.return_value


class TestPingSchema(IsolatedAsyncioTestCase):
    """Test ping schema."""

    async def test_make_model(self):
        ping = Ping(comment="hello", response_requested=True)
        data = ping.serialize()
        model_instance = Ping.deserialize(data)
        assert type(model_instance) is type(ping)



import asyncio
from unittest import IsolatedAsyncioTestCase

from acapy_agent.tests import mock

from ..task_queue import CompletedTask, PendingTask, TaskQueue, task_exc_info


async def retval(val, *, delay=0):
    if delay:
        await asyncio.sleep(delay)
    return val


class TestTaskQueue(IsolatedAsyncioTestCase):
    async def test_run(self):
        queue = TaskQueue()
        task = None
        completed = []

        def done(complete: CompletedTask):
            assert complete.task is task
            assert not complete.exc_info
            completed.append(complete.task.result())

        task = queue.run(retval(1), done)
        assert queue.current_active == 1
        assert len(queue) == queue.current_size == 1
        assert not queue.current_pending
        await queue.flush()
        assert completed == [1]
        assert task.result() == 1

        with self.assertRaises(ValueError):
            queue.run(None, done)

    async def test_put_no_limit(self):
        queue = TaskQueue()
        completed = []

        def done(complete: CompletedTask):
            assert not complete.exc_info
            completed.append(complete.task.result())

        pend = queue.put(retval(1), done)
        assert not queue.pending_tasks
        await queue.flush()
        assert completed == [1]
        assert pend.task.result() == 1

        with self.assertRaises(ValueError):
            queue.put(None, done)

    async def test_put_limited(self):
        queue = TaskQueue(1)
        assert queue.max_active == 1
        assert not queue.cancelled
        completed = set()

        def done(complete: CompletedTask):
            assert not complete.exc_info
            completed.add(complete.task.result())

        pend1 = queue.put(retval(1), done)
        pend2 = queue.put(retval(2), done)
        assert queue.pending_tasks
        await queue.flush()
        assert completed == {1, 2}
        assert pend1.task.result() == 1
        assert pend2.task.result() == 2

    async def test_pending(self):
        coro = retval(1, delay=1)
        pend = PendingTask(coro, None)
        assert str(pend).startswith("<PendingTask")
        task = asyncio.get_event_loop().create_task(coro)
        assert task_exc_info(task) is None
        pend.task = task
        assert pend.task is task
        assert pend.task_future.result() is task
        with self.assertRaises(ValueError):
            pend.task = task
        pend.cancel()
        assert pend.cancelled
        task.cancel()

        with mock.patch.object(pend, "task_future", autospec=True) as mock_future:
            mock_future.cancelled.return_value = True
            pend.task = "a suffusion of yellow"
            mock_future.set_result.assert_not_called()

    async def test_complete(self):
        queue = TaskQueue()
        completed = set()

        def done(complete: CompletedTask):
            assert not complete.exc_info
            completed.add(complete.task.result())

        queue.run(retval(1), done)
        await queue.put(retval(2), done)
        queue.put(retval(3), done)
        await queue.complete()
        assert completed == {1, 2, 3}

        async def noop():
            return

        with (
            mock.patch.object(queue, "drain", mock.MagicMock()) as mock_drain,
            mock.patch.object(queue, "wait_for", mock.CoroutineMock()) as mock_wait_for,
        ):
            mock_drain.side_effect = [queue.loop.create_task(noop()), None]
            await queue.complete(cleanup=True)

    async def test_cancel_pending(self):
        queue = TaskQueue(1)
        completed = set()

        def done(complete: CompletedTask):
            assert not complete.exc_info
            completed.add(complete.task.result())

        queue.run(retval(1), done)
        sleep = queue.run(retval(1, delay=1), done)
        queue.put(retval(2), done)
        queue.put(retval(3), done)
        queue.cancel_pending()
        sleep.cancel()

        await queue.flush()
        assert completed == {1}

    async def test_drain_done(self):
        coro = retval(1, delay=1)
        pend = PendingTask(coro, None)
        queue = TaskQueue(1)
        queue.add_pending(pend)

        with mock.patch.object(
            queue.pending_tasks[0], "task_future", autospec=True
        ) as mock_future:
            mock_future.cancelled.return_value = False
            mock_future.done.return_value = True
            mock_future.set_result.assert_not_called()
            await queue._drain_loop()

    async def test_cancel_all(self):
        queue = TaskQueue(1)
        completed = set()

        def done(complete: CompletedTask):
            assert not complete.exc_info
            completed.add(complete.task.result())

        queue.run(retval(1, delay=1), done)
        queue.put(retval(2), done)
        queue.put(retval(3), done)
        queue.cancel()
        assert queue.cancelled
        await queue.flush()
        assert not completed
        assert not queue.current_size

        co = retval(1)
        with self.assertRaises(RuntimeError):
            queue.run(co, done)
        co.close()

        co = retval(1)
        pend = queue.put(co)
        assert pend.cancelled

    async def test_cancel_long(self):
        queue = TaskQueue()
        task = queue.run(retval(1, delay=5))
        queue.cancel()
        await queue

        # cancellation may take a second
        # assert task.cancelled()

        with self.assertRaises(asyncio.CancelledError):
            await task

    async def test_complete_with_timeout(self):
        queue = TaskQueue()
        task = queue.run(retval(1, delay=5))
        await queue.complete(0.01)

        # cancellation may take a second
        # assert task.cancelled()

        with self.assertRaises(asyncio.CancelledError):
            await task

    async def test_repeat_callback(self):
        # check that running the callback twice does not throw an exception

        queue = TaskQueue()
        completed = []

        def done(complete: CompletedTask):
            assert not complete.exc_info
            completed.append(complete.task.result())

        task = queue.run(retval(1), done)
        await task
        queue.completed_task(task, done, None, {})
        assert completed == [1, 1]

    async def test_timed(self):
        completed = []

        def done(complete: CompletedTask):
            assert not complete.exc_info
            completed.append((complete.task.result(), complete.timing))

        queue = TaskQueue(max_active=1, timed=True, trace_fn=done)
        assert bool(queue)
        task1 = queue.run(retval(1))
        task2 = await queue.put(retval(2))
        assert bool(queue)
        await queue.complete(0.1)

        assert len(completed) == 2
        assert "queued" not in completed[0][1]
        assert "queued" in completed[1][1]



"""Handler for incoming forward invitation messages."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ....out_of_band.v1_0.manager import OutOfBandManager, OutOfBandManagerError
from ....problem_report.v1_0.message import ProblemReport
from ..messages.forward_invitation import ForwardInvitation


class ForwardInvitationHandler(BaseHandler):
    """Handler for incoming forward invitation messages."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler implementation."""
        self._logger.debug("ForwardInvitationHandler called with context %s", context)
        assert isinstance(context.message, ForwardInvitation)

        if not context.connection_ready:
            raise HandlerException(
                "No connection established for forward invitation message"
            )

        # Store invitation
        profile = context.profile
        mgr = OutOfBandManager(profile)

        assert context.message.invitation, "Schema requires invitation in message"
        try:
            await mgr.receive_invitation(context.message.invitation)
        except OutOfBandManagerError as e:
            self._logger.exception("Error receiving forward connection invitation")
            await responder.send_reply(
                ProblemReport(
                    description={
                        "en": e.message,
                        "code": e.error_code or "forward-invitation-error",
                    }
                )
            )



"""Abstract admin server interface."""

from abc import ABC, abstractmethod


class BaseAdminServer(ABC):
    """Admin HTTP server class."""

    @abstractmethod
    async def start(self) -> None:
        """Start the webserver.

        Raises:
            AdminSetupError: If there was an error starting the webserver

        """

    @abstractmethod
    async def stop(self) -> None:
        """Stop the webserver."""



"""Test Data Integrity Cryptosuites."""

from unittest import IsolatedAsyncioTestCase

from acapy_agent.resolver.default.key import KeyDIDResolver
from acapy_agent.resolver.default.web import WebDIDResolver
from acapy_agent.resolver.did_resolver import DIDResolver
from acapy_agent.utils.testing import create_test_profile
from acapy_agent.vc.data_integrity.cryptosuites import EddsaJcs2022
from acapy_agent.vc.data_integrity.models.options import DataIntegrityProofOptions
from acapy_agent.wallet.key_type import KeyTypes
from acapy_agent.wallet.keys.manager import MultikeyManager


class TestEddsaJcs2022(IsolatedAsyncioTestCase):
    """Tests for DI sign and verify."""

    async def asyncSetUp(self):
        self.seed = "00000000000000000000000000000000"
        self.multikey = "z6MkgKA7yrw5kYSiDuQFcye4bMaJpcfHFry3Bx45pdWh3s8i"
        self.verification_method = f"did:key:{self.multikey}#{self.multikey}"
        self.cryptosuite = "eddsa-jcs-2022"
        self.unsecured_document = {"hello": "world"}
        self.options = DataIntegrityProofOptions.deserialize(
            {
                "type": "DataIntegrityProof",
                "cryptosuite": self.cryptosuite,
                "proofPurpose": "assertionMethod",
                "verificationMethod": self.verification_method,
            }
        )

        self.resolver = DIDResolver()
        self.resolver.register_resolver(KeyDIDResolver())
        self.resolver.register_resolver(WebDIDResolver())
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(DIDResolver, self.resolver)
        self.profile.context.injector.bind_instance(KeyTypes, KeyTypes())
        try:
            async with self.profile.session() as session:
                await MultikeyManager(session=session).create(seed=self.seed)
        except Exception:
            pass

    async def test_create_proof(self):
        async with self.profile.session() as session:
            proof = await EddsaJcs2022(session=session).create_proof(
                self.unsecured_document, self.options
            )
            proof = proof.serialize()
            assert isinstance(proof, dict)
            assert proof["type"] == self.options.type
            assert proof["cryptosuite"] == self.options.cryptosuite
            assert proof["proofPurpose"] == self.options.proof_purpose
            assert proof["verificationMethod"] == self.options.verification_method
            assert proof["proofValue"]

    async def test_verify_proof(self):
        async with self.profile.session() as session:
            cryptosuite_instance = EddsaJcs2022(session=session)
            proof = await cryptosuite_instance.create_proof(
                self.unsecured_document, self.options
            )
            secured_document = self.unsecured_document | {"proof": proof.serialize()}
            verification = await cryptosuite_instance.verify_proof(secured_document)
            assert verification.verified



"""v2.0 Presentation exchange record webhook."""


class V20PresExRecordWebhook:
    """Class representing a state only Presentation exchange record webhook."""

    __acceptable_keys_list = [
        "connection_id",
        "pres_ex_id",
        "role",
        "initiator",
        "auto_present",
        "auto_verify",
        "error_msg",
        "thread_id",
        "state",
        "trace",
        "by_format",
        "verified",
        "verified_msgs",
        "created_at",
        "updated_at",
    ]

    def __init__(
        self,
        **kwargs,
    ):
        """Initialize webhook object from V20PresExRecord.

        from a list of accepted attributes.
        """
        [
            self.__setattr__(key, kwargs.get(key))
            for key in self.__acceptable_keys_list
            if kwargs.get(key) is not None
        ]
        if kwargs.get("_id") is not None:
            self.pres_ex_id = kwargs.get("_id")



from unittest import IsolatedAsyncioTestCase

from .....didcomm_prefix import DIDCommPrefix
from .....didexchange.v1_0.message_types import ARIES_PROTOCOL as DIDEX_1_1
from ...messages.invitation import InvitationMessage
from ..invitation import InvitationRecord

TEST_DID = "did:sov:55GkHamhTU1ZbTbV2ab9DE"


class TestInvitationRecord(IsolatedAsyncioTestCase):
    def test_invitation_record(self):
        """Test invitation record."""
        invi_rec = InvitationRecord(invi_msg_id="12345")
        assert invi_rec.invitation_id is None  # not saved
        assert isinstance(invi_rec, InvitationRecord)
        assert invi_rec.record_value == {
            "invitation_url": None,
            "state": None,
            "trace": False,
            "oob_id": None,
        }

        another = InvitationRecord(invi_msg_id="99999")
        assert invi_rec != another


class TestInvitationRecordSchema(IsolatedAsyncioTestCase):
    def test_make_record(self):
        """Test making record."""
        invi = InvitationMessage(
            label="A label",
            handshake_protocols=[DIDCommPrefix.qualify_current(DIDEX_1_1)],
            services=[TEST_DID],
        )
        data = {
            "invitation_id": "0",
            "state": InvitationRecord.STATE_AWAIT_RESPONSE,
            "invitation": invi.serialize(),
        }
        model_instance = InvitationRecord.deserialize(data)
        assert isinstance(model_instance, InvitationRecord)

        assert data.items() <= model_instance.serialize().items()

        model_instance = InvitationRecord(
            invitation_id="0",
            state=InvitationRecord.STATE_AWAIT_RESPONSE,
        )
        model_instance.invitation = invi  # exercise setter
        assert data.items() <= model_instance.serialize().items()
        model_instance = InvitationRecord(
            invitation_id="0",
            state=InvitationRecord.STATE_AWAIT_RESPONSE,
            invitation=invi.serialize(),
        )
        assert data.items() <= model_instance.serialize().items()



from unittest import IsolatedAsyncioTestCase

from ...config.injection_context import InjectionContext
from ...ledger.base import BaseLedger
from ...ledger.multiple_ledger.base_manager import BaseMultipleLedgerManager
from ...tests import mock
from ...utils.testing import create_test_profile
from .. import indy_tails_server as test_module

TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
CRED_DEF_ID = f"{TEST_DID}:3:CL:1234:default"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:0"


class TestIndyTailsServer(IsolatedAsyncioTestCase):
    async def test_upload_no_tails_upload_url_x(self):
        context = InjectionContext(settings={"ledger.genesis_transactions": "dummy"})
        indy_tails = test_module.IndyTailsServer()

        with self.assertRaises(test_module.TailsServerNotConfiguredError):
            await indy_tails.upload_tails_file(context, REV_REG_ID, "/tmp/dummy/path")

    async def test_upload(self):
        context = InjectionContext(
            settings={
                "ledger.genesis_transactions": "dummy",
                "tails_server_upload_url": "http://1.2.3.4:8088",
            }
        )
        indy_tails = test_module.IndyTailsServer()

        with mock.patch.object(test_module, "put_file", mock.CoroutineMock()) as mock_put:
            mock_put.return_value = "tails-hash"
            (ok, text) = await indy_tails.upload_tails_file(
                context,
                REV_REG_ID,
                "/tmp/dummy/path",
            )
            assert ok
            assert text == context.settings["tails_server_upload_url"] + "/" + REV_REG_ID

    async def test_upload_indy_vdr(self):
        self.profile = await create_test_profile()
        self.profile.settings["tails_server_upload_url"] = "http://1.2.3.4:8088"
        mock_multi_ledger_manager = mock.MagicMock(
            BaseMultipleLedgerManager, autospec=True
        )
        mock_multi_ledger_manager.get_write_ledgers = mock.CoroutineMock(
            return_value=[
                "test_ledger_id_1",
                "test_ledger_id_2",
            ]
        )
        self.profile.context.injector.bind_instance(
            BaseMultipleLedgerManager, mock_multi_ledger_manager
        )
        mock_ledger = mock.MagicMock(BaseLedger, autospec=True)
        mock_ledger.pool = mock.MagicMock(
            genesis_txns="dummy genesis transactions",
        )
        self.profile.context.injector.bind_instance(BaseLedger, mock_ledger)
        indy_tails = test_module.IndyTailsServer()

        with mock.patch.object(test_module, "put_file", mock.CoroutineMock()) as mock_put:
            mock_put.return_value = "tails-hash"
            (ok, text) = await indy_tails.upload_tails_file(
                self.profile.context,
                REV_REG_ID,
                "/tmp/dummy/path",
            )
            assert ok
            assert (
                text
                == self.profile.settings["tails_server_upload_url"] + "/" + REV_REG_ID
            )

    async def test_upload_x(self):
        context = InjectionContext(
            settings={
                "ledger.genesis_transactions": "dummy",
                "tails_server_upload_url": "http://1.2.3.4:8088",
            }
        )
        indy_tails = test_module.IndyTailsServer()

        with mock.patch.object(test_module, "put_file", mock.CoroutineMock()) as mock_put:
            mock_put.side_effect = test_module.PutError("Server down for maintenance")

            (ok, text) = await indy_tails.upload_tails_file(
                context, REV_REG_ID, "/tmp/dummy/path"
            )
            assert not ok
            assert text == "Server down for maintenance"



"""Author specific for indy legacy."""

from typing import Optional

from aiohttp import web

from acapy_agent.connections.models.conn_record import ConnRecord
from acapy_agent.core.profile import Profile
from acapy_agent.messaging.models.base import BaseModelError
from acapy_agent.protocols.endorse_transaction.v1_0.util import get_endorser_connection_id
from acapy_agent.storage.error import StorageNotFoundError


async def get_endorser_info(
    profile: Profile, options: Optional[dict] = None
) -> tuple[str, str]:
    """Gets the endorser did for the current transaction."""
    options = options or {}
    endorser_connection_id = options.get("endorser_connection_id", None)
    if not endorser_connection_id:
        endorser_connection_id = await get_endorser_connection_id(profile)

    if not endorser_connection_id:
        raise web.HTTPForbidden(reason="No endorser connection found")

    try:
        async with profile.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(
                session, endorser_connection_id
            )
            endorser_info = await connection_record.metadata_get(session, "endorser_info")
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(
            reason=f"Connection for endorser with id {endorser_connection_id} not found"
        ) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not endorser_info:
        raise web.HTTPForbidden(
            reason=(
                "Endorser Info is not set up in "
                "connection metadata for this connection record"
            )
        )
    if "endorser_did" not in endorser_info.keys():
        raise web.HTTPForbidden(
            reason=(
                '"endorser_did" is not set in "endorser_info" '
                "in connection metadata for this connection record"
            )
        )

    return endorser_info["endorser_did"], endorser_connection_id



from unittest import IsolatedAsyncioTestCase

from .....did.did_key import DIDKey
from .....utils.testing import create_test_profile
from .....wallet.base import BaseWallet
from .....wallet.key_type import ED25519
from ....tests.data import (
    TEST_LD_DOCUMENT,
    TEST_LD_DOCUMENT_BAD_SIGNED_ED25519,
    TEST_LD_DOCUMENT_SIGNED_ED25519,
    TEST_VC_DOCUMENT,
    TEST_VC_DOCUMENT_SIGNED_ED25519,
)
from ....tests.document_loader import custom_document_loader
from ...crypto.wallet_key_pair import WalletKeyPair
from ...ld_proofs import sign, verify
from ...purposes.assertion_proof_purpose import AssertionProofPurpose
from ..ed25519_signature_2018 import Ed25519Signature2018


class TestEd25519Signature2018(IsolatedAsyncioTestCase):
    test_seed = "testseed000000000000000000000001"

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            self.key = await wallet.create_signing_key(
                key_type=ED25519, seed=self.test_seed
            )
        self.verification_method = DIDKey.from_public_key_b58(
            self.key.verkey, ED25519
        ).key_id

        self.sign_key_pair = WalletKeyPair(
            profile=self.profile,
            key_type=ED25519,
            public_key_base58=self.key.verkey,
        )
        self.verify_key_pair = WalletKeyPair(profile=self.profile, key_type=ED25519)

    async def test_sign_ld_proofs(self):
        signed = await sign(
            document=TEST_LD_DOCUMENT,
            suite=Ed25519Signature2018(
                key_pair=self.sign_key_pair,
                verification_method=self.verification_method,
            ),
            document_loader=custom_document_loader,
            purpose=AssertionProofPurpose(),
        )

        assert signed

    async def test_verify_ld_proofs(self):
        result = await verify(
            document=TEST_LD_DOCUMENT_SIGNED_ED25519,
            suites=[Ed25519Signature2018(key_pair=self.verify_key_pair)],
            document_loader=custom_document_loader,
            purpose=AssertionProofPurpose(),
        )

        assert result
        assert result.verified

    async def test_verify_ld_proofs_not_verified_bad_signature(self):
        result = await verify(
            document=TEST_LD_DOCUMENT_BAD_SIGNED_ED25519,
            suites=[Ed25519Signature2018(key_pair=self.verify_key_pair)],
            document_loader=custom_document_loader,
            purpose=AssertionProofPurpose(),
        )

        assert result
        assert not result.verified

    async def test_verify_ld_proofs_not_verified_unsigned_statement(self):
        MODIFIED_DOCUMENT = {
            **TEST_LD_DOCUMENT_SIGNED_ED25519,
            "unsigned_claim": "oops",
        }
        result = await verify(
            document=MODIFIED_DOCUMENT,
            suites=[Ed25519Signature2018(key_pair=self.verify_key_pair)],
            document_loader=custom_document_loader,
            purpose=AssertionProofPurpose(),
        )

        assert result
        assert not result.verified

    async def test_verify_ld_proofs_not_verified_changed_statement(self):
        MODIFIED_DOCUMENT = {
            **TEST_LD_DOCUMENT_SIGNED_ED25519,
            "email": "someOtherEmail@example.com",
        }
        result = await verify(
            document=MODIFIED_DOCUMENT,
            suites=[Ed25519Signature2018(key_pair=self.verify_key_pair)],
            document_loader=custom_document_loader,
            purpose=AssertionProofPurpose(),
        )

        assert result
        assert not result.verified

    async def test_sign_vc(self):
        signed = await sign(
            document=TEST_VC_DOCUMENT,
            suite=Ed25519Signature2018(
                key_pair=self.sign_key_pair,
                verification_method=self.verification_method,
            ),
            document_loader=custom_document_loader,
            purpose=AssertionProofPurpose(),
        )

        assert signed

    async def test_verify_vc(self):
        result = await verify(
            document=TEST_VC_DOCUMENT_SIGNED_ED25519,
            suites=[Ed25519Signature2018(key_pair=self.verify_key_pair)],
            document_loader=custom_document_loader,
            purpose=AssertionProofPurpose(),
        )

        assert result
        assert result.verified



from unittest import IsolatedAsyncioTestCase

from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....indy.issuer import IndyIssuer
from ....ledger.base import BaseLedger
from ....ledger.multiple_ledger.ledger_requests_executor import IndyLedgerRequestsExecutor
from ....multitenant.base import BaseMultitenantManager
from ....multitenant.manager import MultitenantManager
from ....storage.base import BaseStorage
from ....tests import mock
from ....utils.testing import create_test_profile
from .. import routes as test_module

SCHEMA_ID = "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0"
CRED_DEF_ID = "WgWxqztrNooG92RXvxSTWv:3:CL:20:tag"


class TestCredentialDefinitionRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.profile_injector = self.profile.context.injector

        self.ledger = mock.create_autospec(BaseLedger)
        self.ledger.__aenter__ = mock.CoroutineMock(return_value=self.ledger)
        self.ledger.create_and_send_credential_definition = mock.CoroutineMock(
            return_value=(
                CRED_DEF_ID,
                {"cred": "def", "signed_txn": "..."},
                True,
            )
        )
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value={"cred": "def", "signed_txn": "..."}
        )
        self.profile_injector.bind_instance(BaseLedger, self.ledger)

        self.issuer = mock.create_autospec(IndyIssuer)
        self.profile_injector.bind_instance(IndyIssuer, self.issuer)

        self.storage = mock.create_autospec(BaseStorage)
        self.storage.find_all_records = mock.CoroutineMock(
            return_value=[mock.MagicMock(value=CRED_DEF_ID)]
        )
        self.session_inject[BaseStorage] = self.storage

        self.context = AdminRequestContext.test_context(
            self.session_inject, profile=self.profile
        )
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_send_credential_definition(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.request.query = {"create_transaction_for_endorser": "false"}

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.credential_definitions_send_credential_definition(
                self.request
            )
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {
                    "sent": {"credential_definition_id": CRED_DEF_ID},
                    "credential_definition_id": CRED_DEF_ID,
                }
            )

    async def test_send_credential_definition_create_transaction_for_endorser(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
            mock.patch.object(
                test_module.web, "json_response", mock.MagicMock()
            ) as mock_response,
        ):
            mock_txn_mgr.return_value = mock.MagicMock(
                create_record=mock.CoroutineMock(
                    return_value=mock.MagicMock(
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    )
                )
            )
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_did": ("did"),
                        "endorser_name": ("name"),
                    }
                )
            )
            result = await test_module.credential_definitions_send_credential_definition(
                self.request
            )
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {
                    "sent": {"credential_definition_id": CRED_DEF_ID},
                    "txn": {"...": "..."},
                }
            )

    async def test_send_credential_definition_create_transaction_for_endorser_storage_x(
        self,
    ):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with (
            mock.patch.object(
                ConnRecord, "retrieve_by_id", mock.CoroutineMock()
            ) as mock_conn_rec_retrieve,
            mock.patch.object(
                test_module, "TransactionManager", mock.MagicMock()
            ) as mock_txn_mgr,
        ):
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_did": ("did"),
                        "endorser_name": ("name"),
                    }
                )
            )
            mock_txn_mgr.return_value = mock.MagicMock(
                create_record=mock.CoroutineMock(side_effect=test_module.StorageError())
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_definitions_send_credential_definition(
                    self.request
                )

    async def test_send_credential_definition_create_transaction_for_endorser_not_found_x(
        self,
    ):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_definitions_send_credential_definition(
                    self.request
                )

    async def test_send_credential_definition_create_transaction_for_endorser_base_model_x(
        self,
    ):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.side_effect = test_module.BaseModelError()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_definitions_send_credential_definition(
                    self.request
                )

    async def test_send_credential_definition_create_transaction_for_endorser_no_endorser_info_x(
        self,
    ):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(return_value=None)
            )
            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_definitions_send_credential_definition(
                    self.request
                )

    async def test_send_credential_definition_create_transaction_for_endorser_no_endorser_did_x(
        self,
    ):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.request.query = {
            "create_transaction_for_endorser": "true",
            "conn_id": "dummy",
        }

        with mock.patch.object(
            ConnRecord, "retrieve_by_id", mock.CoroutineMock()
        ) as mock_conn_rec_retrieve:
            mock_conn_rec_retrieve.return_value = mock.MagicMock(
                metadata_get=mock.CoroutineMock(
                    return_value={
                        "endorser_name": ("name"),
                    }
                )
            )
            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_definitions_send_credential_definition(
                    self.request
                )

    async def test_send_credential_definition_no_ledger(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.context.injector.clear_binding(BaseLedger)
        self.profile_injector.clear_binding(BaseLedger)
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.credential_definitions_send_credential_definition(
                self.request
            )

    async def test_send_credential_definition_ledger_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "schema_id": "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0",
                "support_revocation": False,
                "tag": "tag",
            }
        )

        self.request.query = {"create_transaction_for_endorser": "false"}

        self.ledger.__aenter__ = mock.CoroutineMock(
            side_effect=test_module.LedgerError("oops")
        )
        with self.assertRaises(test_module.web.HTTPBadRequest):
            await test_module.credential_definitions_send_credential_definition(
                self.request
            )

    async def test_created(self):
        self.request.match_info = {"cred_def_id": CRED_DEF_ID}

        with mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.credential_definitions_created(self.request)
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {"credential_definition_ids": [CRED_DEF_ID]}
            )

    async def test_get_credential_definition(self):
        mock_ledger_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_ledger_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile_injector.bind_instance(
            IndyLedgerRequestsExecutor, mock_ledger_executor
        )
        self.request.match_info = {"cred_def_id": CRED_DEF_ID}
        with mock.patch.object(test_module.web, "json_response") as mock_response:
            result = await test_module.credential_definitions_get_credential_definition(
                self.request
            )
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                    "credential_definition": {"cred": "def", "signed_txn": "..."},
                }
            )

    async def test_get_credential_definition_multitenant(self):
        self.profile_injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        self.request.match_info = {"cred_def_id": CRED_DEF_ID}
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            result = await test_module.credential_definitions_get_credential_definition(
                self.request
            )
            assert result == mock_response.return_value
            mock_response.assert_called_once_with(
                {
                    "ledger_id": "test_ledger_id",
                    "credential_definition": {"cred": "def", "signed_txn": "..."},
                }
            )

    async def test_get_credential_definition_no_ledger(self):
        mock_ledger_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_ledger_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, None)
        )
        self.profile_injector.bind_instance(
            IndyLedgerRequestsExecutor, mock_ledger_executor
        )

        self.request.match_info = {"cred_def_id": CRED_DEF_ID}
        self.context.injector.clear_binding(BaseLedger)
        self.profile_injector.clear_binding(BaseLedger)
        with self.assertRaises(test_module.web.HTTPForbidden):
            await test_module.credential_definitions_get_credential_definition(
                self.request
            )

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



from unittest import IsolatedAsyncioTestCase

from .....admin.request_context import AdminRequestContext
from .....core.event_bus import EventBus, MockEventBus
from .....utils.testing import create_test_profile
from .. import util as test_module
from ..models.menu_form import MenuForm
from ..models.menu_form_param import MenuFormParam
from ..models.menu_option import MenuOption


class TestActionMenuUtil(IsolatedAsyncioTestCase):
    async def test_save_retrieve_delete_connection_menu(self):
        context = AdminRequestContext.test_context({}, await create_test_profile())

        mock_event_bus = MockEventBus()
        context.profile.context.injector.bind_instance(EventBus, mock_event_bus)

        menu = test_module.Menu(
            title="title",
            description="description",
            errormsg=None,
            options=[
                MenuOption(
                    name=f"name-{i}",
                    title=f"option-title-{i}",
                    description=f"option-description-{i}",
                    disabled=bool(i % 2),
                    form=MenuForm(
                        title=f"form-title-{i}",
                        description=f"form-description-{i}",
                        params=[
                            MenuFormParam(
                                name=f"param-name-{i}-{j}",
                                title=f"param-title-{i}-{j}",
                                default=f"param-default-{i}-{j}",
                                description=f"param-description-{i}-{j}",
                                input_type="int",
                                required=bool((i + j) % 2),
                            )
                            for j in range(2)
                        ],
                    ),
                )
                for i in range(3)
            ],
        )
        connection_id = "connid"

        for _ in range(2):  # once to add, once to update
            await test_module.save_connection_menu(menu, connection_id, context)

            assert len(mock_event_bus.events) == 1
            (_, event) = mock_event_bus.events[0]
            assert event.topic == "acapy::actionmenu::received"
            assert event.payload["connection_id"] == connection_id
            assert event.payload["menu"] == menu.serialize()
            mock_event_bus.events.clear()

        # retrieve connection menu
        assert (
            await test_module.retrieve_connection_menu(connection_id, context)
        ).serialize() == menu.serialize()

        # delete connection menu
        await test_module.save_connection_menu(None, connection_id, context)

        assert len(mock_event_bus.events) == 1
        (_, event) = mock_event_bus.events[0]
        assert event.topic == "acapy::actionmenu::received"
        assert event.payload == {"connection_id": connection_id, "menu": None}
        mock_event_bus.events.clear()

        # retrieve no menu
        assert await test_module.retrieve_connection_menu(connection_id, context) is None



"""HTTP Universal DID Resolver."""

import logging
import re
from typing import Iterable, Optional, Pattern, Sequence, Text, Union

import aiohttp

from ...config.injection_context import InjectionContext
from ...core.profile import Profile
from ..base import BaseDIDResolver, DIDNotFound, ResolverError, ResolverType

LOGGER = logging.getLogger(__name__)
DEFAULT_ENDPOINT = "https://dev.uniresolver.io/1.0"


def _compile_supported_did_regex(patterns: Iterable[Union[str, Pattern]]):
    """Create regex from list of regex."""
    return re.compile(
        "(?:"
        + "|".join(
            [
                pattern.pattern if isinstance(pattern, Pattern) else pattern
                for pattern in patterns
            ]
        )
        + ")"
    )


class UniversalResolver(BaseDIDResolver):
    """Universal DID Resolver with HTTP bindings."""

    def __init__(
        self,
        *,
        endpoint: Optional[str] = None,
        supported_did_regex: Optional[Pattern] = None,
        bearer_token: Optional[str] = None,
    ):
        """Initialize UniversalResolver."""
        super().__init__(ResolverType.NON_NATIVE)
        self._endpoint = endpoint
        self._supported_did_regex = supported_did_regex

        self.__default_headers = (
            {"Authorization": f"Bearer {bearer_token}"} if bearer_token else {}
        )

    async def setup(self, context: InjectionContext):
        """Perform setup, populate supported method list, configuration."""

        # configure endpoint
        endpoint = context.settings.get_str("resolver.universal")
        if endpoint == "DEFAULT" or not endpoint:
            endpoint = DEFAULT_ENDPOINT
        self._endpoint = endpoint

        # configure authorization
        token = context.settings.get_str("resolver.universal.token")
        self.__default_headers = {"Authorization": f"Bearer {token}"} if token else {}

        # configure supported methods
        supported = context.settings.get("resolver.universal.supported")
        if supported is None:
            supported_did_regex = await self._get_supported_did_regex()
        else:
            supported_did_regex = _compile_supported_did_regex(supported)

        self._supported_did_regex = supported_did_regex

    @property
    def supported_did_regex(self) -> Pattern:
        """Return supported methods regex."""
        if not self._supported_did_regex:
            raise ResolverError("Resolver has not been set up")

        return self._supported_did_regex

    async def _resolve(
        self,
        _profile: Profile,
        did: str,
        service_accept: Optional[Sequence[Text]] = None,
    ) -> dict:
        """Resolve DID through remote universal resolver."""

        async with aiohttp.ClientSession(headers=self.__default_headers) as session:
            async with session.get(f"{self._endpoint}/identifiers/{did}") as resp:
                if resp.status == 200:
                    doc = await resp.json()
                    did_doc = doc["didDocument"]
                    LOGGER.info("Retrieved doc: %s", did_doc)
                    return did_doc
                if resp.status == 404:
                    raise DIDNotFound(f"{did} not found by {self.__class__.__name__}")

                text = await resp.text()
                raise ResolverError(
                    f"Unexpected status from universal resolver ({resp.status}): {text}"
                )

    async def _fetch_resolver_props(self) -> dict:
        """Retrieve universal resolver properties."""
        async with aiohttp.ClientSession(headers=self.__default_headers) as session:
            try:
                async with session.get(f"{self._endpoint}/properties/") as resp:
                    if 200 <= resp.status < 400:
                        return await resp.json()
                    raise ResolverError(
                        "Failed to retrieve resolver properties: " + await resp.text()
                    )
            except aiohttp.ClientError as err:
                raise ResolverError(f"Failed to fetch resolver properties: {err}")

    async def _get_supported_did_regex(self):
        props = await self._fetch_resolver_props()

        def _get_patterns():
            """Handle both old and new properties responses."""
            patterns = list(props.values())[0].get("http", {}).get("pattern")
            if not patterns:
                return props.keys()
            else:
                return [driver["http"]["pattern"] for driver in props.values()]

        patterns = _get_patterns()
        return _compile_supported_did_regex(patterns)



"""Authentication proof purpose class."""

from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Optional

from ..document_loader import DocumentLoaderMethod
from ..error import LinkedDataProofException
from ..validation_result import PurposeResult
from .controller_proof_purpose import ControllerProofPurpose

# Avoid circular dependency
if TYPE_CHECKING:
    from ..suites import LinkedDataProof


class AuthenticationProofPurpose(ControllerProofPurpose):
    """Authentication proof purpose."""

    term = "authentication"

    def __init__(
        self,
        *,
        challenge: str,
        domain: Optional[str] = None,
        date: Optional[datetime] = None,
        max_timestamp_delta: Optional[timedelta] = None,
    ):
        """Initialize new AuthenticationProofPurpose instance."""
        super().__init__(
            term=AuthenticationProofPurpose.term,
            date=date,
            max_timestamp_delta=max_timestamp_delta,
        )

        self.challenge = challenge
        self.domain = domain

    def validate(
        self,
        *,
        proof: dict,
        document: dict,
        suite: "LinkedDataProof",
        verification_method: dict,
        document_loader: DocumentLoaderMethod,
    ) -> PurposeResult:
        """Validate whether challenge and domain are valid."""
        try:
            if proof.get("challenge") != self.challenge:
                raise LinkedDataProofException(
                    f"The challenge is not as expected; challenge="
                    f"{proof.get('challenge')}, expected={self.challenge}"
                )

            if self.domain and (proof.get("domain") != self.domain):
                raise LinkedDataProofException(
                    f"The domain is not as expected; "
                    f"domain={proof.get('domain')}, expected={self.domain}"
                )

            return super().validate(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=document_loader,
            )
        except Exception as e:
            return PurposeResult(valid=False, error=str(e))

    def update(self, proof: dict) -> dict:
        """Update poof purpose, challenge and domain on proof."""
        proof = super().update(proof)
        proof["challenge"] = self.challenge

        if self.domain:
            proof["domain"] = self.domain

        return proof

    def __eq__(self, o: object) -> bool:
        """Check if object is same as ProofPurpose."""
        if isinstance(o, AuthenticationProofPurpose):
            return (
                super().__eq__(o)
                and self.challenge == o.challenge
                and self.domain == o.domain
            )

        return False



"""Record for out of band invitations."""

import json
from typing import Any, Mapping, Optional, Union

from marshmallow import fields, validate

from .....connections.models.conn_record import ConnRecord
from .....core.profile import ProfileSession
from .....messaging.decorators.service_decorator import (
    ServiceDecorator,
    ServiceDecoratorSchema,
)
from .....messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from .....messaging.valid import UUID4_EXAMPLE
from .....storage.base import BaseStorage
from .....storage.error import StorageNotFoundError
from .....storage.record import StorageRecord
from ..messages.invitation import InvitationMessage, InvitationMessageSchema


class OobRecord(BaseExchangeRecord):
    """Represents an out of band record."""

    class Meta:
        """OobRecord metadata."""

        schema_class = "OobRecordSchema"

    RECORD_TYPE = "oob_record"
    RECORD_TYPE_METADATA = ConnRecord.RECORD_TYPE_METADATA
    RECORD_ID_NAME = "oob_id"
    RECORD_TOPIC = "out_of_band"
    TAG_NAMES = {
        "invi_msg_id",
        "attach_thread_id",
        "our_recipient_key",
        "connection_id",
        "reuse_msg_id",
    }

    STATE_INITIAL = "initial"
    STATE_PREPARE_RESPONSE = "prepare-response"
    STATE_AWAIT_RESPONSE = "await-response"
    STATE_NOT_ACCEPTED = "reuse-not-accepted"
    STATE_ACCEPTED = "reuse-accepted"
    STATE_DONE = "done"

    ROLE_SENDER = "sender"
    ROLE_RECEIVER = "receiver"

    def __init__(
        self,
        *,
        state: str,
        invi_msg_id: str,
        role: str,
        invitation: Union[InvitationMessage, Mapping[str, Any]],
        their_service: Optional[Union[ServiceDecorator, Mapping[str, Any]]] = None,
        connection_id: Optional[str] = None,
        reuse_msg_id: Optional[str] = None,
        oob_id: Optional[str] = None,
        attach_thread_id: Optional[str] = None,
        our_recipient_key: Optional[str] = None,
        our_service: Optional[Union[ServiceDecorator, Mapping[str, Any]]] = None,
        multi_use: bool = False,
        trace: bool = False,
        **kwargs,
    ):
        """Initialize a new OobRecord."""
        super().__init__(oob_id, state, trace=trace, **kwargs)
        self._id = oob_id
        self.state = state
        self.invi_msg_id = invi_msg_id
        self.role = role
        self._invitation = InvitationMessage.serde(invitation)
        self.connection_id = connection_id
        self.reuse_msg_id = reuse_msg_id
        self._their_service = ServiceDecorator.serde(their_service)
        self._our_service = ServiceDecorator.serde(our_service)
        self.attach_thread_id = attach_thread_id
        self.our_recipient_key = our_recipient_key
        self.multi_use = multi_use
        self.trace = trace

    @property
    def oob_id(self) -> str:
        """Accessor for the ID associated with this exchange."""
        return self._id

    @property
    def invitation(self) -> Optional[InvitationMessage]:
        """Accessor; get deserialized view."""
        return None if self._invitation is None else self._invitation.de

    @invitation.setter
    def invitation(self, value):
        """Setter; store de/serialized views."""
        self._invitation = InvitationMessage.serde(value)

    @property
    def our_service(self) -> Optional[ServiceDecorator]:
        """Accessor; get deserialized view."""
        return None if self._our_service is None else self._our_service.de

    @our_service.setter
    def our_service(self, value: Union[ServiceDecorator, Mapping[str, Any]]):
        """Setter; store de/serialized views."""
        self._our_service = ServiceDecorator.serde(value)

    @property
    def their_service(self) -> Optional[ServiceDecorator]:
        """Accessor; get deserialized view."""
        return None if self._their_service is None else self._their_service.de

    @their_service.setter
    def their_service(self, value: Union[ServiceDecorator, Mapping[str, Any]]):
        """Setter; store de/serialized vies."""
        self._their_service = ServiceDecorator.serde(value)

    @property
    def record_value(self) -> dict:
        """Accessor for the JSON record value generated for this invitation."""
        return {
            **{
                prop: getattr(self, prop)
                for prop in (
                    "state",
                    "their_service",
                    "connection_id",
                    "role",
                    "invi_msg_id",
                    "multi_use",
                )
            },
            **{
                prop: getattr(self, f"_{prop}").ser
                for prop in ("invitation", "our_service", "their_service")
                if getattr(self, prop) is not None
            },
        }

    async def delete_record(self, session: ProfileSession):
        """Perform connection record deletion actions.

        Args:
            session (ProfileSession): session

        """
        await super().delete_record(session)

        # Delete metadata
        if self.connection_id:
            storage = session.inject(BaseStorage)
            await storage.delete_all_records(
                self.RECORD_TYPE_METADATA,
                {"connection_id": self.connection_id},
            )

    async def metadata_get(
        self, session: ProfileSession, key: str, default: Optional[Any] = None
    ) -> Any:
        """Retrieve arbitrary metadata associated with this connection.

        Args:
            session (ProfileSession): session used for storage
            key (str): key identifying metadata
            default (Any): default value to get; type should be a JSON
                compatible value.

        Returns:
            Any: metadata stored by key

        """
        assert self.connection_id
        storage: BaseStorage = session.inject(BaseStorage)
        try:
            record = await storage.find_record(
                self.RECORD_TYPE_METADATA,
                {"key": key, "connection_id": self.connection_id},
            )
            return json.loads(record.value)
        except StorageNotFoundError:
            return default

    async def metadata_set(self, session: ProfileSession, key: str, value: Any):
        """Set arbitrary metadata associated with this connection.

        Args:
            session (ProfileSession): session used for storage
            key (str): key identifying metadata
            value (Any): value to set
        """
        assert self.connection_id
        value = json.dumps(value)
        storage: BaseStorage = session.inject(BaseStorage)
        try:
            record = await storage.find_record(
                self.RECORD_TYPE_METADATA,
                {"key": key, "connection_id": self.connection_id},
            )
            await storage.update_record(record, value, record.tags)
        except StorageNotFoundError:
            record = StorageRecord(
                self.RECORD_TYPE_METADATA,
                value,
                {"key": key, "connection_id": self.connection_id},
            )
            await storage.add_record(record)

    async def metadata_delete(self, session: ProfileSession, key: str):
        """Delete custom metadata associated with this connection.

        Args:
            session (ProfileSession): session used for storage
            key (str): key of metadata to delete
        """
        assert self.connection_id
        storage: BaseStorage = session.inject(BaseStorage)
        try:
            record = await storage.find_record(
                self.RECORD_TYPE_METADATA,
                {"key": key, "connection_id": self.connection_id},
            )
            await storage.delete_record(record)
        except StorageNotFoundError as err:
            raise KeyError(f"{key} not found in connection metadata") from err

    async def metadata_get_all(self, session: ProfileSession) -> dict:
        """Return all custom metadata associated with this connection.

        Args:
            session (ProfileSession): session used for storage

        Returns:
            dict: dictionary representation of all metadata values

        """
        assert self.connection_id
        storage: BaseStorage = session.inject(BaseStorage)
        records = await storage.find_all_records(
            self.RECORD_TYPE_METADATA,
            {"connection_id": self.connection_id},
        )
        return {record.tags["key"]: json.loads(record.value) for record in records}

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class OobRecordSchema(BaseExchangeSchema):
    """Schema to allow serialization/deserialization of invitation records."""

    class Meta:
        """OobRecordSchema metadata."""

        model_class = OobRecord

    oob_id = fields.Str(
        required=True,
        metadata={"description": "Oob record identifier", "example": UUID4_EXAMPLE},
    )
    state = fields.Str(
        required=True,
        validate=validate.OneOf(
            OobRecord.get_attributes_by_prefix("STATE_", walk_mro=True)
        ),
        metadata={
            "description": "Out of band message exchange state",
            "example": OobRecord.STATE_AWAIT_RESPONSE,
        },
    )
    invi_msg_id = fields.Str(
        required=True,
        metadata={
            "description": "Invitation message identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    invitation = fields.Nested(
        InvitationMessageSchema(),
        required=True,
        metadata={"description": "Out of band invitation message"},
    )

    their_service = fields.Nested(ServiceDecoratorSchema(), required=False)

    connection_id = fields.Str(
        required=False,
        metadata={
            "description": "Connection record identifier",
            "example": UUID4_EXAMPLE,
        },
    )

    attach_thread_id = fields.Str(
        required=False,
        metadata={
            "description": "Connection record identifier",
            "example": UUID4_EXAMPLE,
        },
    )

    our_recipient_key = fields.Str(
        required=False,
        metadata={
            "description": "Recipient key used for oob invitation",
            "example": UUID4_EXAMPLE,
        },
    )

    role = fields.Str(
        required=False,
        validate=validate.OneOf(
            OobRecord.get_attributes_by_prefix("ROLE_", walk_mro=False)
        ),
        metadata={"description": "OOB Role", "example": OobRecord.ROLE_RECEIVER},
    )

    multi_use = fields.Boolean(
        required=False,
        metadata={
            "description": "Allow for multiple uses of the oob invitation",
            "example": True,
        },
    )
